from distutils.version import LooseVersion
import numbers
import operator
import warnings
import operator
import json

import numpy as np
import pandas as pd


from pandas.core.arrays import ExtensionArray
from pandas.core.dtypes.dtypes import ExtensionDtype
from pandas.api.extensions import ExtensionArray

from collections.abc import Iterable
from arcgis.geometry import Geometry, Point, Polygon, Polyline

# -----------------------------------------------------------------------------
# pandas version checker
# -----------------------------------------------------------------------------
PANDAS_GE_024 = str(pd.__version__) >= LooseVersion("0.24.0")
PANDAS_GE_025 = str(pd.__version__) >= LooseVersion("0.25.0")
PANDAS_GE_10 = str(pd.__version__) >= LooseVersion("1")


# --------------------------------------------------------------------------
def _isna(value):
    """
    Check if scalar value is NA-like (None or np.nan).

    Custom version that only works for scalars (returning True or False),
    as `pd.isna` also works for array-like input returning a boolean array.
    """
    if value is None:
        return True
    elif isinstance(value, float) and np.isnan(value):
        return True
    else:
        return False


# --------------------------------------------------------------------------
def _unary_geo(op, left, *args, **kwargs):
    """
    Unary operation that returns new geometries

    **used for accessing properties on objects**

    :return: GeoArray
    """
    data = np.empty(len(left), dtype=object)
    data[:] = [getattr(geom, op, None) for geom in left]
    return GeoArray(data)


# --------------------------------------------------------------------------
def _unary_op(op, left, null_value=False):
    """
    Unary operation that returns a Series

    **used for accessing properties on objects**

    :return: pd.Series

    """
    data = np.empty(len(left), dtype=object)
    data[:] = [getattr(geom, op, null_value) for geom in left]
    return data


# --------------------------------------------------------------------------
def _binary_predicate(name, left, right, *args, **kwargs):
    """

    Binary operation performed on the GeoArray that returns only boolean ndarray.

    Supports:
    -  contains
    -  disjoint
    -  intersect
    -  touches
    -  crosses
    -  within
    -  overlaps
    -  equals

    Parameters
    ----------

         name: string
         left : GeoArray
         right: GeoArray or Geometry

    :return: np.array (should be dtype bool)

    """
    if isinstance(right, pd.Series):
        right = right.values
    if isinstance(right, Geometry):
        data = np.empty(len(left), dtype=bool)
        data[:] = [
            getattr(s, name)(right, *args, **kwargs) if s is not None else left.na_value
            for s in left
        ]
        return data
    elif isinstance(right, GeoArray):
        if len(left) != len(right):
            msg = "Lengths of inputs do not match. Left: {0}, Right: {1}".format(
                len(left), len(right)
            )
            raise ValueError(msg)
        data = np.empty(len(left), dtype=bool)
        data[:] = [
            getattr(this_elem, name)(other_elem, *args, **kwargs)
            if not (this_elem is None or other_elem is None)
            else False
            for this_elem, other_elem in zip(left, right)
        ]
        return data
    else:
        raise TypeError("Type not known: {0} vs {1}".format(type(left), type(right)))


# --------------------------------------------------------------------------
def _binary_op(name, left, right=None, *args, **kwargs):
    """Binary operation on GeoArray that returns a ndarray of dtype object"""

    if isinstance(right, pd.Series):
        right = right.values
    null_value = None
    if right is None:
        data = np.empty(len(left), dtype=object)
        data[:] = [
            getattr(s, name)(*args, **kwargs) if s is not None else null_value
            for s in left
        ]
        return data
    elif isinstance(right, Geometry):
        data = np.empty(len(left), dtype=object)
        data[:] = [
            getattr(s, name)(right, *args, **kwargs) if s is not None else null_value
            for s in left
        ]
        return data
    elif isinstance(right, GeoArray):
        if len(left) != len(right):
            msg = "Lengths of inputs do not match. Left: {0}, Right: {1}".format(
                len(left), len(right)
            )
            raise ValueError(msg)
        data = np.empty(len(left), dtype=object)
        data[:] = [
            getattr(this_elem, name)(other_elem, *args, **kwargs)
            if not (this_elem is None or other_elem is None)
            else null_value
            for this_elem, other_elem in zip(left, right)
        ]
        return data
    else:
        raise TypeError("Type not known: {0} vs {1}".format(type(left), type(right)))


# --------------------------------------------------------------------------
def _binary_op_geo(name, left, right=None, *args, **kwargs):
    """Binary operation on GeoArray that returns a GeoArray"""

    if isinstance(right, pd.Series):
        right = right.values
    null_value = None
    if right is None:
        data = np.empty(len(left), dtype=object)
        data[:] = [
            getattr(s, name)(*args, **kwargs) if s is not None else null_value
            for s in left
        ]
        return GeoArray(data)
    elif isinstance(right, Geometry):
        data = np.empty(len(left), dtype=object)
        data[:] = [
            getattr(s, name)(right, *args, **kwargs) if s is not None else null_value
            for s in left
        ]
        return GeoArray(data)
    elif isinstance(right, GeoArray):
        if len(left) != len(right):
            msg = "Lengths of inputs do not match. Left: {0}, Right: {1}".format(
                len(left), len(right)
            )
            raise ValueError(msg)
        data = np.empty(len(left), dtype=object)
        data[:] = [
            getattr(this_elem, name)(other_elem, *args, **kwargs)
            if not (this_elem is None or other_elem is None)
            else null_value
            for this_elem, other_elem in zip(left, right)
        ]
        return GeoArray(data)
    else:
        raise TypeError("Type not known: {0} vs {1}".format(type(left), type(right)))


# --------------------------------------------------------------------------


class GeoType(ExtensionDtype):
    type = Geometry
    name = "geometry"
    na_value = None
    # np.nan

    @classmethod
    def construct_from_string(cls, string):
        if string == cls.name:
            return cls()
        else:
            raise TypeError(
                "Cannot construct a '{}' from '{}'".format(cls.__name__, string)
            )

    @classmethod
    def construct_array_type(cls):
        return GeoArray


if PANDAS_GE_024:
    from pandas.api.extensions import register_extension_dtype

    register_extension_dtype(GeoType)


class GeoArray(ExtensionArray):
    """
    Class wrapping a numpy array of Shapely objects and
    holding the array-based implementations.
    """

    _dtype = GeoType()

    def __init__(self, values):
        if isinstance(values, self.__class__):
            data = values.data
        elif isinstance(values, pd.Series):
            data = values.values
        elif isinstance(values, (list, tuple)):
            data = np.array(values)
        elif isinstance(values, np.ndarray):
            data = values
        elif not isinstance(values, np.ndarray):
            raise TypeError("'data' should be array of geometry objects.")
        elif not values.ndim == 1:
            raise ValueError(
                "'data' should be a 1-dimensional array of geometry objects."
            )
        self.data = data
        self._validate_data()

    def _validate_data(self):
        data = self.data
        check = np.where(self.data != None)[0]
        if len(check) > 0:
            vindx = check[0]
            if isinstance(data[vindx], Geometry) == False:
                self.data[:] = [Geometry(d) if d else None for d in data]

                # Extra step for shapely, need to transform to correct Geometry type instance if not already
                geom = self.data[0]
                if (
                    not isinstance(geom, Point)
                    or not isinstance(geom, Polyline)
                    or not isinstance(geom, Polygon)
                ):
                    if "type" in geom and (
                        geom["type"] == "Point" or geom["type"] == "MultiPoint"
                    ):
                        self.data[:] = [Point(d) if d else None for d in data]
                    elif "type" in geom and geom["type"] == "Polyline":
                        self.data[:] = [Polyline(d) if d else None for d in data]
                    elif "type" in geom and (
                        geom["type"] == "Polygon" or geom["type"] == "MultiPolygon"
                    ):
                        self.data[:] = [Polygon(d) if d else None for d in data]

    def __arrow_array__(self, type=None):
        """converts the data to a pyarrow array"""
        import pyarrow

        return pyarrow.array([d.WKB for d in self.data if d], type=type)

    def __eq__(self, other: Geometry):
        """Checks if the Geometries are Equal"""
        if isinstance(other, Geometry):
            return self.equals(other)
        elif isinstance(other, GeoArray):
            return np.array_equal(self, other)
        else:
            raise ValueError(
                "Input must be a arcgis.geometry.Geometry or arcgis.features.geo.GeoArray"
            )

    def __ne__(self, other: Geometry):
        """Checks if the Geometries are Equal"""
        if isinstance(other, Geometry):
            return self.equals(other) == False
        else:
            raise ValueError("Input must be a arcgis.geometry.Geometry")

    def _formatting_values_backport(self):
        return np.array(self._format_values(), dtype="object")

    def _format_values(self):
        if self.data.ndim == 0:
            return ""
        return [_format(x) if x else None for x in self.data]

    @property
    def dtype(self):
        return self._dtype

    def __len__(self):
        return self.shape[0]

    @classmethod
    def from_geometry(cls, data, copy=False):
        """ """
        if copy:
            data = data.copy()
        new = GeoArray([])
        new.data = np.array(data)
        return new

    def __getitem__(self, idx):
        if isinstance(idx, numbers.Integral):
            return self.data[idx]
        # array-like, slice
        if PANDAS_GE_10 and pd.api.types.is_list_like(idx):
            # for pandas >= 1.0, validate and convert IntegerArray/BooleanArray
            # to numpy array
            if not pd.api.types.is_array_like(idx):
                idx = pd.array(idx)
            dtype = idx.dtype
            if pd.api.types.is_bool_dtype(dtype):
                idx = pd.api.indexers.check_array_indexer(self, idx)
            elif pd.api.types.is_integer_dtype(dtype):
                idx = np.asarray(idx, dtype="int")
        if isinstance(idx, (Iterable, slice)):
            return GeoArray(self.data[idx])
        else:
            raise TypeError("Index type not supported", idx)

    def __setitem__(self, key, value):
        if isinstance(value, pd.Series):
            value = value.values
        if isinstance(value, GeoArray):
            if isinstance(key, numbers.Integral):
                raise ValueError("cannot set a single element with an array")
            self.data[key] = value.data
        elif isinstance(value, Geometry) or _isna(value):
            if _isna(value):
                # internally only use None as missing value indicator
                # but accept others
                value = None
            if isinstance(key, (list, np.ndarray)):
                value_array = np.empty(1, dtype=object)
                value_array[:] = [value]
                self.data[key] = value_array
            else:
                self.data[key] = value
        elif isinstance(value, str) and value != "":
            value = Geometry(value)
            self.data[key] = value
        else:
            raise TypeError(
                "Value should be either a Geometry or None, got %s" % str(value)
            )

    # -------------------------------------------------------------------------
    # general array like compat
    # -------------------------------------------------------------------------

    @property
    def size(self):
        return self.data.size

    @property
    def shape(self):
        return (self.size,)

    @property
    def ndim(self):
        return len(self.shape)

    def copy(self, *args, **kwargs):
        # still taking args/kwargs for compat with pandas 0.24
        return GeoArray(self.data.copy())

    def take(self, indices, allow_fill=False, fill_value=None):
        from pandas.api.extensions import take

        if allow_fill:
            if fill_value is None or pd.isna(fill_value):
                fill_value = 0

        result = take(self.data, indices, allow_fill=allow_fill, fill_value=fill_value)
        if fill_value == 0:
            result[result == 0] = None
        return GeoArray(result)

    def _fill(self, idx, value):
        """Fill index locations with value

        Value should be a Geometry
        """
        if not (isinstance(value, Geometry) or value is None):
            raise TypeError(
                "Value should be either a Geometry or None, got %s" % str(value)
            )
        # self.data[idx] = value
        self.data[idx] = np.array([value], dtype=object)
        return self

    def fillna(self, value=None, method=None, limit=None):
        """Fill NA/NaN values using the specified method.

        Parameters
        ----------
        value : scalar, array-like
            If a scalar value is passed it is used to fill all missing values.
            Alternatively, an array-like 'value' can be given. It's expected
            that the array-like have the same length as 'self'.
        method : {'backfill', 'bfill', 'pad', 'ffill', None}, default None
            Method to use for filling holes in reindexed Series
            pad / ffill: propagate last valid observation forward to next valid
            backfill / bfill: use NEXT valid observation to fill gap
        limit : int, default None
            If method is specified, this is the maximum number of consecutive
            NaN values to forward/backward fill. In other words, if there is
            a gap with more than this number of consecutive NaNs, it will only
            be partially filled. If method is not specified, this is the
            maximum number of entries along the entire axis where NaNs will be
            filled.

        Returns
        -------
        filled : ExtensionArray with NA/NaN filled
        """
        if method is not None:
            raise NotImplementedError("fillna with a method is not yet supported")

        if _isna(value):
            value = None
        elif not isinstance(value, Geometry):
            raise NotImplementedError(
                "fillna currently only supports filling with a scalar geometry"
            )

        mask = self.isna()
        new_values = self.copy()

        if mask.any():
            # fill with value
            new_values = new_values._fill(mask, value)

        return new_values

    def astype(self, dtype, copy=True):
        """
        Cast to a NumPy array with 'dtype'.

        Parameters
        ----------
        dtype : str or dtype
            Typecode or data-type to which the array is cast.
        copy : bool, default True
            Whether to copy the data, even if not necessary. If False,
            a copy is made only if the old dtype does not match the
            new dtype.

        Returns
        -------
        array : ndarray
            NumPy ndarray with 'dtype' for its dtype.
        """
        if isinstance(dtype, GeoType):
            if copy:
                return self.copy()
            else:
                return self
        elif pd.api.types.is_string_dtype(dtype) and not pd.api.types.is_object_dtype(
            dtype
        ):
            return np.array([g.JSON for g in self.data])
        else:
            return np.array(self, dtype=dtype, copy=copy)

    @property
    def na_value(self):
        return self.dtype.na_value

    def isna(self):
        """
        Boolean NumPy array indicating if each value is missing
        """
        return np.array([g is self.na_value for g in self.data], dtype="bool")

    def unique(self):
        """Compute the ExtensionArray of unique values.

        Returns
        -------
        uniques : ExtensionArray
        """
        from pandas import factorize

        _, uniques = factorize(self)
        return uniques

    @property
    def nbytes(self):
        return self.data.nbytes

    # -------------------------------------------------------------------------
    # ExtensionArray specific
    # -------------------------------------------------------------------------

    @classmethod
    def _from_sequence(cls, scalars, dtype=None, copy=False):
        """
        Construct a new ExtensionArray from a sequence of scalars.

        Parameters
        ----------
        scalars : Sequence
            Each element will be an instance of the scalar type for this
            array, ``cls.dtype.type``.
        dtype : dtype, optional
            Construct for this particular dtype. This should be a Dtype
            compatible with the ExtensionArray.
        copy : boolean, default False
            If True, copy the underlying data.

        Returns
        -------
        ExtensionArray
        """
        data = np.empty(len(scalars), dtype=object)
        data[:] = [Geometry(s) for s in scalars]
        return cls(data)

    def _values_for_factorize(self):
        """Return an array and missing value suitable for factorization.

        Returns
        -------
        values : ndarray
            An array suitable for factoraization. This should maintain order
            and be a supported dtype (Float64, Int64, UInt64, String, Object).
            By default, the extension array is cast to object dtype.
        na_value : object
            The value in `values` to consider missing. This will be treated
            as NA in the factorization routines, so it will be coded as
            `na_sentinal` and not included in `uniques`. By default,
            ``np.nan`` is used.
        """
        return self, 0

    @classmethod
    def _from_factorized(cls, values):
        """
        Reconstruct an ExtensionArray after factorization.

        Parameters
        ----------
        values : ndarray
            An integer ndarray with the factorized values.

        See Also
        --------
        pandas.factorize
        ExtensionArray.factorize
        """
        return cls(values)

    def _values_for_argsort(self):
        # type: () -> np.ndarray
        """Return values for sorting.

        Returns
        -------
        ndarray
            The transformed values should maintain the ordering between values
            within the array.

        See Also
        --------
        ExtensionArray.argsort
        """
        # Note: this is used in `ExtensionArray.argsort`.
        raise TypeError("geometries are not orderable")

    def _formatter(self, boxed=False):
        """Formatting function for scalar values.

        This is used in the default '__repr__'. The returned formatting
        function receives instances of your scalar type.

        Parameters
        ----------
        boxed: bool, default False
            An indicated for whether or not your array is being printed
            within a Series, DataFrame, or Index (True), or just by
            itself (False). This may be useful if you want scalar values
            to appear differently within a Series versus on its own (e.g.
            quoted or not).

        Returns
        -------
        Callable[[Any], str]
            A callable that gets instances of the scalar type and
            returns a string. By default, :func:`repr` is used
            when ``boxed=False`` and :func:`str` is used when
            ``boxed=True``.
        """
        if boxed:
            return json.dumps
        return repr

    @classmethod
    def _concat_same_type(cls, to_concat):
        """
        Concatenate multiple array

        Parameters
        ----------
        to_concat : sequence of this type

        Returns
        -------
        ExtensionArray
        """
        data = np.concatenate([ga.data for ga in to_concat])
        return GeoArray(data)

    def _reduce(self, name, skipna=True, **kwargs):
        # including the base class version here (that raises by default)
        # because this was not yet defined in pandas 0.23
        if name == "any" or name == "all":
            return getattr(self.data, name)()
        raise TypeError(
            "cannot perform {name} with type {dtype}".format(
                name=name, dtype=self.dtype
            )
        )

    def __array__(self, dtype=None):
        """
        The numpy array interface.

        Returns
        -------
        values : numpy array
        """
        return self.data

    # ----------------------------------------------------------------------
    @property
    def area(self):
        """returns the geometry area"""
        return _unary_op("area", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def as_arcpy(self):
        """returns the geometry area"""
        return _unary_op("as_arcpy", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def as_shapely(self):
        """returns the geometry area"""
        return _unary_op("as_shapely", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def centroid(self):
        """returns Geometry centroid"""
        return _unary_op("centroid", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def extent(self):
        """returns the extent of the geometry"""
        return _unary_op("extent", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def first_point(self):
        """
        The first coordinate point of the geometry for each entry.
        """
        return _unary_geo("first_point", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def geoextent(self):
        return _unary_op("geoextent", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def geometry_type(self):
        return _unary_op("geometry_type", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def has_z(self):
        return _unary_op("has_z", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def has_m(self):
        return _unary_op("has_m", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def hull_rectangle(self):
        return _unary_op("hull_rectangle", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def is_empty(self):
        return _unary_op("is_empty", self.data, False)

    # ----------------------------------------------------------------------
    @property
    def is_multipart(self):
        return _unary_op("is_multipart", self.data, False)

    # ----------------------------------------------------------------------
    @property
    def is_valid(self):
        return _binary_op(name="is_valid", left=self.data, right=None)

    # ----------------------------------------------------------------------
    @property
    def JSON(self):
        return _unary_op("JSON", self.data, "")

    # ----------------------------------------------------------------------
    @property
    def label_point(self):
        return _unary_geo("label_point", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def last_point(self):
        return _unary_geo("last_point", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def length(self):
        return _unary_op("length", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def length3D(self):
        return _unary_op("length3D", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def part_count(self):
        return _unary_op("part_count", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def point_count(self):
        return _unary_op("point_count", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def spatial_reference(self):
        return _unary_op("spatial_reference", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def true_centroid(self):
        return _unary_geo("true_centroid", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def WKB(self):
        return _unary_op("WKB", self.data, None)

    # ----------------------------------------------------------------------
    @property
    def WKT(self):
        return _unary_op("WKT", self.data, None)

    # ----------------------------------------------------------------------
    def angle_distance_to(self, second_geometry, method="GEODESIC"):
        """
        Returns a tuple of angle and distance to another point using a
        measurement type.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required Geometry.  A arcgis.Geometry object.
        ---------------     --------------------------------------------------------------------
        method              Optional String. PLANAR measurements reflect the projection of geographic
                            data onto the 2D surface (in other words, they will not take into
                            account the curvature of the earth). GEODESIC, GREAT_ELLIPTIC, and
                            LOXODROME measurement types may be chosen as an alternative, if desired.
        ===============     ====================================================================

        :return: a tuple of angle and distance to another point using a measurement type.
        """
        return _binary_op(
            name="angle_distance_to",
            left=self.data,
            right=second_geometry,
            **{"method": method},
        )

    # ----------------------------------------------------------------------
    def boundary(self):
        """
        Constructs the boundary of the geometry.

        :return: arcgis.geometry.Polyline
        """

        return _binary_op_geo(name="boundary", left=self.data, right=None)

    # ----------------------------------------------------------------------
    def buffer(self, distance):
        """
        Constructs a polygon at a specified distance from the geometry.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        distance            Required float. The buffer distance. The buffer distance is in the
                            same units as the geometry that is being buffered.
                            A negative distance can only be specified against a polygon geometry.
        ===============     ====================================================================

        :return: arcgis.geometry.Polygon
        """
        return _binary_op_geo(name="buffer", left=self.data, **{"distance": distance})

    # ----------------------------------------------------------------------
    def clip(self, envelope):
        """
        Constructs the intersection of the geometry and the specified extent.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        envelope            required tuple. The tuple must have (XMin, YMin, XMax, YMax) each value
                            represents the lower left bound and upper right bound of the extent.
        ===============     ====================================================================

        :return: output geometry clipped to extent

        """
        return _binary_op_geo(name="clip", left=self.data, **{"envelope": envelope})

    # ----------------------------------------------------------------------
    def contains(self, second_geometry, relation=None):
        """
        Indicates if the base geometry contains the comparison geometry.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        relation            Optional string. The spatial relationship type.

                            + BOUNDARY - Relationship has no restrictions for interiors or boundaries.
                            + CLEMENTINI - Interiors of geometries must intersect. Specifying CLEMENTINI is equivalent to specifying None. This is the default.
                            + PROPER - Boundaries of geometries must not intersect.
        ===============     ====================================================================

        :return: boolean
        """
        return _binary_predicate(
            name="contains",
            left=self.data,
            right=second_geometry,
            **{"relation": relation},
        )

    # ----------------------------------------------------------------------
    def convex_hull(self):
        """
        Constructs the geometry that is the minimal bounding polygon such
        that all outer angles are convex.
        """
        return _binary_op_geo(name="convex_hull", left=self.data)

    # ----------------------------------------------------------------------
    def crosses(self, second_geometry):
        """
        Indicates if the two geometries intersect in a geometry of a lesser
        shape type.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: boolean

        """
        return _binary_predicate(name="crosses", left=self.data, right=second_geometry)

    # ----------------------------------------------------------------------
    def cut(self, cutter):
        """
        Splits this geometry into a part left of the cutting polyline, and
        a part right of it.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        cutter              Required Polyline. The cuttin polyline geometry
        ===============     ====================================================================

        :return: a list of two geometries

        """
        return _binary_op_geo(name="cut", left=self.data, right=cutter)

    # ----------------------------------------------------------------------
    def densify(self, method, distance, deviation):
        """
        Creates a new geometry with added vertices

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        method              Required String. The type of densification, DISTANCE, ANGLE, or GEODESIC
        ---------------     --------------------------------------------------------------------
        distance            Required float. The maximum distance between vertices. The actual
                            distance between vertices will usually be less than the maximum
                            distance as new vertices will be evenly distributed along the
                            original segment. If using a type of DISTANCE or ANGLE, the
                            distance is measured in the units of the geometry's spatial
                            reference. If using a type of GEODESIC, the distance is measured
                            in meters.
        ---------------     --------------------------------------------------------------------
        deviation           Required float. Densify uses straight lines to approximate curves.
                            You use deviation to control the accuracy of this approximation.
                            The deviation is the maximum distance between the new segment and
                            the original curve. The smaller its value, the more segments will
                            be required to approximate the curve.
        ===============     ====================================================================

        :return: arcgis.geometry.Geometry

        """
        return _binary_op_geo(
            name="densify",
            left=self.data,
            **{
                "method": method,
                "distance": distance,
                "deviation": deviation,
            },
        )

    # ----------------------------------------------------------------------
    def difference(self, second_geometry):
        """
        Constructs the geometry that is composed only of the region unique
        to the base geometry but not part of the other geometry. The
        following illustration shows the results when the red polygon is the
        source geometry.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: arcgis.geometry.Geometry

        """
        return _binary_op_geo(name="difference", left=self.data, right=second_geometry)

    # ----------------------------------------------------------------------
    def disjoint(self, second_geometry):
        """
        Indicates if the base and comparison geometries share no points in
        common.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: boolean

        """
        return _binary_predicate(
            name="disjoint", left=self.data, right=second_geometry, **{}
        )

    # ----------------------------------------------------------------------
    def distance_to(self, second_geometry):
        """
        Returns the minimum distance between two geometries. If the
        geometries intersect, the minimum distance is 0.
        Both geometries must have the same projection.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: float

        """
        return _binary_op(
            name="distance_to", left=self.data, right=second_geometry, **{}
        )

    # ----------------------------------------------------------------------
    def equals(self, second_geometry):
        """
        Indicates if the base and comparison geometries are of the same
        shape type and define the same set of points in the plane. This is
        a 2D comparison only; M and Z values are ignored.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: boolean


        """
        return _binary_predicate(name="equals", left=self.data, right=second_geometry)

    # ----------------------------------------------------------------------
    def generalize(self, max_offset):
        """
        Creates a new simplified geometry using a specified maximum offset
        tolerance.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        max_offset          Required float. The maximum offset tolerance.
        ===============     ====================================================================

        :return: arcgis.geometry.Geometry

        """
        return _binary_op_geo(
            name="generalize", left=self.data, **{"max_offset": max_offset}
        )

    # ----------------------------------------------------------------------
    def get_area(self, method, units=None):
        """
        Returns the area of the feature using a measurement type.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        method              Required String. PLANAR measurements reflect the projection of
                            geographic data onto the 2D surface (in other words, they will not
                            take into account the curvature of the earth). GEODESIC,
                            GREAT_ELLIPTIC, LOXODROME, and PRESERVE_SHAPE measurement types
                            may be chosen as an alternative, if desired.
        ---------------     --------------------------------------------------------------------
        units               Optional String. Areal unit of measure keywords: ACRES | ARES | HECTARES
                            | SQUARECENTIMETERS | SQUAREDECIMETERS | SQUAREINCHES | SQUAREFEET
                            | SQUAREKILOMETERS | SQUAREMETERS | SQUAREMILES |
                            SQUAREMILLIMETERS | SQUAREYARDS
        ===============     ====================================================================

        :return: float

        """
        return _binary_op(
            name="get_area",
            left=self.data,
            **{"method": method, "units": units},
        )

    # ----------------------------------------------------------------------
    def get_length(self, method, units):
        """
        Returns the length of the feature using a measurement type.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        method              Required String. PLANAR measurements reflect the projection of
                            geographic data onto the 2D surface (in other words, they will not
                            take into account the curvature of the earth). GEODESIC,
                            GREAT_ELLIPTIC, LOXODROME, and PRESERVE_SHAPE measurement types
                            may be chosen as an alternative, if desired.
        ---------------     --------------------------------------------------------------------
        units               Required String. Linear unit of measure keywords: CENTIMETERS |
                            DECIMETERS | FEET | INCHES | KILOMETERS | METERS | MILES |
                            MILLIMETERS | NAUTICALMILES | YARDS
        ===============     ====================================================================

        :return: float

        """
        return _binary_op(
            name="get_length",
            left=self.data,
            **{"method": method, "units": units},
        )

    # ----------------------------------------------------------------------
    def get_part(self, index=None):
        """
        Returns an array of point objects for a particular part of geometry
        or an array containing a number of arrays, one for each part.

        **requires arcpy**

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        index               Required Integer. The index position of the geometry.
        ===============     ====================================================================

        :return: arcpy.Array

        """
        return _binary_op(name="get_part", left=self.data, **{"index": index})

    # ----------------------------------------------------------------------
    def intersect(self, second_geometry, dimension=1):
        """
        Constructs a geometry that is the geometric intersection of the two
        input geometries. Different dimension values can be used to create
        different shape types. The intersection of two geometries of the
        same shape type is a geometry containing only the regions of overlap
        between the original geometries.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        dimension           Required Integer. The topological dimension (shape type) of the
                            resulting geometry.

                            + 1  -A zero-dimensional geometry (point or multipoint).
                            + 2  -A one-dimensional geometry (polyline).
                            + 4  -A two-dimensional geometry (polygon).

        ===============     ====================================================================

        :return: boolean array

        """
        return _binary_predicate(
            name="intersect",
            left=self.data,
            right=second_geometry,
            **{"dimension": dimension},
        )

    # ----------------------------------------------------------------------
    def measure_on_line(self, second_geometry, as_percentage=False):
        """
        Returns a measure from the start point of this line to the in_point.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        as_percentage       Optional Boolean. If False, the measure will be returned as a
                            distance; if True, the measure will be returned as a percentage.
        ===============     ====================================================================

        :return: float

        """
        return _binary_op(
            name="measure_on_line",
            left=self.data,
            right=second_geometry,
            **{"as_percentage": as_percentage},
        )

    # ----------------------------------------------------------------------
    def overlaps(self, second_geometry):
        """
        Indicates if the intersection of the two geometries has the same
        shape type as one of the input geometries and is not equivalent to
        either of the input geometries.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: boolean

        """
        return _binary_predicate(name="overlaps", left=self.data, right=second_geometry)

    # ----------------------------------------------------------------------
    def point_from_angle_and_distance(self, angle, distance, method="GEODESCIC"):
        """
        Returns a point at a given angle and distance in degrees and meters
        using the specified measurement type.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        angle               Required Float. The angle in degrees to the returned point.
        ---------------     --------------------------------------------------------------------
        distance            Required Float. The distance in meters to the returned point.
        ---------------     --------------------------------------------------------------------
        method              Optional String. PLANAR measurements reflect the projection of geographic
                            data onto the 2D surface (in other words, they will not take into
                            account the curvature of the earth). GEODESIC, GREAT_ELLIPTIC,
                            LOXODROME, and PRESERVE_SHAPE measurement types may be chosen as
                            an alternative, if desired.
        ===============     ====================================================================

        :return: arcgis.geometry.Geometry


        """
        return _binary_op_geo(
            name="point_from_angle_and_distance",
            left=self.data,
            **{"angle": angle, "distance": distance, "method": method},
        )

    # ----------------------------------------------------------------------
    def position_along_line(self, value, use_percentage=False):
        """
        Returns a point on a line at a specified distance from the beginning
        of the line.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        value               Required Float. The distance along the line.
        ---------------     --------------------------------------------------------------------
        use_percentage      Optional Boolean. The distance may be specified as a fixed unit
                            of measure or a ratio of the length of the line. If True, value
                            is used as a percentage; if False, value is used as a distance.
                            For percentages, the value should be expressed as a double from
                            0.0 (0%) to 1.0 (100%).
        ===============     ====================================================================

        :return: arcgis.gis.Geometry

        """
        return _binary_op_geo(
            name="position_along_line",
            left=self.data,
            **{"value": value, "use_percentage": use_percentage},
        )

    # ----------------------------------------------------------------------
    def project_as(self, spatial_reference, transformation_name=None):
        """
        Projects a geometry and optionally applies a geotransformation.

        ====================     ====================================================================
        **Parameter**             **Description**
        --------------------     --------------------------------------------------------------------
        spatial_reference        Required SpatialReference. The new spatial reference. This can be a
                                 SpatialReference object or the coordinate system name.
        --------------------     --------------------------------------------------------------------
        transformation_name      Required String. The geotransformation name.
        ====================     ====================================================================

        :return: arcgis.geometry.Geometry
        """
        return _binary_op_geo(
            name="project_as",
            left=self.data,
            **{
                "spatial_reference": spatial_reference,
                "transformation_name": transformation_name,
            },
        )

    # ----------------------------------------------------------------------
    def query_point_and_distance(self, second_geometry, use_percentage=False):
        """
        Finds the point on the polyline nearest to the in_point and the
        distance between those points. Also returns information about the
        side of the line the in_point is on as well as the distance along
        the line where the nearest point occurs.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        as_percentage       Optional boolean - if False, the measure will be returned as
                            distance, True, measure will be a percentage
        ===============     ====================================================================

        :return: tuple

        """
        return _binary_op(
            name="query_point_and_distance",
            left=self.data,
            right=second_geometry,
            **{"use_percentage": use_percentage},
        )

    # ----------------------------------------------------------------------
    def segment_along_line(self, start_measure, end_measure, use_percentage=False):
        """
        Returns a Polyline between start and end measures. Similar to
        Polyline.positionAlongLine but will return a polyline segment between
        two points on the polyline instead of a single point.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        start_measure       Required Float. The starting distance from the beginning of the line.
        ---------------     --------------------------------------------------------------------
        end_measure         Required Float. The ending distance from the beginning of the line.
        ---------------     --------------------------------------------------------------------
        use_percentage      Optional Boolean. The start and end measures may be specified as
                            fixed units or as a ratio.
                            If True, start_measure and end_measure are used as a percentage; if
                            False, start_measure and end_measure are used as a distance. For
                            percentages, the measures should be expressed as a double from 0.0
                            (0 percent) to 1.0 (100 percent).
        ===============     ====================================================================

        :return: Geometry

        """
        return _binary_op_geo(
            name="segment_along_line",
            left=self.data,
            **{
                "start_measure": start_measure,
                "end_measure": end_measure,
                "use_percentage": use_percentage,
            },
        )

    # ----------------------------------------------------------------------
    def snap_to_line(self, second_geometry):
        """
        Returns a new point based on in_point snapped to this geometry.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: arcgis.gis.Geometry

        """
        return _binary_op_geo(
            name="snap_to_line", left=self.data, right=second_geometry
        )

    # ----------------------------------------------------------------------
    def symmetric_difference(self, second_geometry):
        """
        Constructs the geometry that is the union of two geometries minus the
        instersection of those geometries.

        The two input geometries must be the same shape type.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: arcgis.gis.Geometry
        """
        return _binary_op_geo(
            name="symmetric_difference",
            left=self.data,
            right=second_geometry,
        )

    # ----------------------------------------------------------------------
    def touches(self, second_geometry):
        """
        Indicates if the boundaries of the geometries intersect.


        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: boolean
        """
        return _binary_predicate(name="touches", left=self.data, right=second_geometry)

    # ----------------------------------------------------------------------
    def union(self, second_geometry):
        """
        Constructs the geometry that is the set-theoretic union of the input
        geometries.


        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: arcgis.gis.Geometry
        """
        return _binary_op_geo(name="union", left=self.data, right=second_geometry)

    # ----------------------------------------------------------------------
    def within(self, second_geometry, relation=None):
        """
        Indicates if the base geometry is within the comparison geometry.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        relation            Optional String. The spatial relationship type.

                            - BOUNDARY  - Relationship has no restrictions for interiors or boundaries.
                            - CLEMENTINI  - Interiors of geometries must intersect. Specifying CLEMENTINI is equivalent to specifying None. This is the default.
                            - PROPER  - Boundaries of geometries must not intersect.

        ===============     ====================================================================

        :return: boolean

        """
        return _binary_predicate(
            name="within",
            left=self.data,
            right=second_geometry,
            **{"relation": relation},
        )


def _format(g):
    if g in {None, np.nan}:
        return ""
    return json.dumps(g)



# Copyright 2020 The MMSegmentation Authors.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


_base_ = "./ocrnet_base.py"
norm_cfg = dict(type="BN", requires_grad=True)
model = dict(
    pretrained="open-mmlab://msra/hrnetv2_w48",
    backbone=dict(
        extra=dict(
            stage2=dict(num_channels=(48, 96)),
            stage3=dict(num_channels=(48, 96, 192)),
            stage4=dict(num_channels=(48, 96, 192, 384)),
        )
    ),
    decode_head=[
        dict(
            type="FCNHead",
            in_channels=[48, 96, 192, 384],
            channels=sum([48, 96, 192, 384]),
            input_transform="resize_concat",
            in_index=(0, 1, 2, 3),
            kernel_size=1,
            num_convs=1,
            norm_cfg=norm_cfg,
            concat_input=False,
            dropout_ratio=-1,
            num_classes=19,
            align_corners=False,
            loss_decode=dict(
                type="CrossEntropyLoss", use_sigmoid=False, loss_weight=0.4
            ),
        ),
        dict(
            type="OCRHead",
            in_channels=[48, 96, 192, 384],
            channels=512,
            ocr_channels=256,
            input_transform="resize_concat",
            in_index=(0, 1, 2, 3),
            norm_cfg=norm_cfg,
            dropout_ratio=-1,
            num_classes=19,
            align_corners=False,
            loss_decode=dict(
                type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0
            ),
        ),
    ],
)

checkpoint = "https://download.openmmlab.com/mmsegmentation/v0.5/ocrnet/ocrnet_hr48_512x1024_160k_cityscapes/ocrnet_hr48_512x1024_160k_cityscapes_20200602_191037-dfbf1b0c.pth"



from ._lazy import LazyLoader

warnings = LazyLoader("warnings")
contextlib = LazyLoader("contextlib")
requests = LazyLoader("requests")
urllib3 = LazyLoader("urllib3")

_HISTORIC_ENVIRONMENTAL_SETTING = requests.Session.merge_environment_settings


@contextlib.contextmanager
def no_ssl_verification():
    opened_adapters = set()

    def merge_environment_settings(self, url, proxies, stream, verify, cert):
        # Verification happens only once per connection so we need to close
        # all the opened adapters once we're done. Otherwise, the effects of
        # verify=False persist beyond the end of this context manager.

        opened_adapters.add(self.get_adapter(url))

        settings = _HISTORIC_ENVIRONMENTAL_SETTING(
            self, url, proxies, stream, verify, cert
        )
        settings["verify"] = False

        return settings

    requests.Session.merge_environment_settings = merge_environment_settings

    try:
        requests.packages.urllib3.disable_warnings(
            category=urllib3.exceptions.InsecureRequestWarning
        )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", urllib3.exceptions.InsecureRequestWarning)
            yield
    finally:
        requests.Session.merge_environment_settings = _HISTORIC_ENVIRONMENTAL_SETTING
        warnings.simplefilter("default", urllib3.exceptions.InsecureRequestWarning)
        for adapter in opened_adapters:
            try:
                adapter.close()
            except:  # pragma: no cover
                pass



import torch
import matplotlib.pyplot as plt
import math
import numpy as np
from .common import get_nbatches, image_batch_stretcher
from .._utils.env import is_arcgispronotebook


def show_batch_labeled_tiles(self, rows=3, **kwargs):  # parameters adjusted in kwargs
    """
    This function randomly picks a few training chips and visualizes them.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    rows                    Optional Integer.
                            Number of rows to display.
                            Default: 3.
    ---------------------   -------------------------------------------
    alpha                   Optional Float.
                            Opacity of the lables for the corresponding
                            images. Values range between 0 and 1, where
                            1 means opaque.
    -------------------------------------------------------------------

    """
    from .._utils.common import denorm_x

    nrows = rows
    ncols = kwargs.get("ncols", nrows)
    # start_index = kwargs.get('start_index', 0) # Does not work with dataloader

    # Modify nrows and ncols according to the dataset
    n_items = kwargs.get("n_items", nrows * ncols)
    n_items = min(n_items, len(self.x))
    nrows = math.ceil(n_items / ncols)
    nbatches = math.ceil(n_items / self.batch_size)

    type_data_loader = kwargs.get(
        "data_loader", "training"
    )  # options : traininig, validation, testing
    if type_data_loader == "training":
        data_loader = self.train_dl
    elif type_data_loader == "validation":
        data_loader = self.valid_dl
    elif type_data_loader == "testing":
        data_loader = self.test_dl
    else:
        e = Exception(
            f"could not find {type_data_loader} in data. Please ensure that the data loader type is traininig, validation or testing "
        )
        raise (e)

    rgb_bands = kwargs.get("rgb_bands", self._symbology_rgb_bands)
    nodata = kwargs.get("nodata", 0)
    imsize = kwargs.get("imsize", 5)
    statistics_type = kwargs.get(
        "statistics_type", "dataset"
    )  # Accepted Values `dataset`, `DRA`
    stretch_type = kwargs.get(
        "stretch_type", "minmax"
    )  # Accepted Values `minmax`, `percentclip`

    e = Exception(
        "`rgb_bands` should be a valid band_order, list or tuple of length 3 or 1."
    )
    symbology_bands = []
    if not (len(rgb_bands) == 3 or len(rgb_bands) == 1):
        raise (e)
    for b in rgb_bands:
        if type(b) == str:
            b_index = self._bands.index(b)
        elif type(b) == int:
            self._bands[
                b
            ]  # To check if the band index specified by the user really exists.
            b_index = b
        else:
            raise (e)
        b_index = self._extract_bands.index(b_index)
        symbology_bands.append(b_index)

    # Get Batch
    x_batch, y_batch = get_nbatches(data_loader, nbatches)
    x_batch = torch.cat(x_batch)
    y_batch = torch.cat(y_batch)

    # Denormalize X
    x_batch = denorm_x(x_batch, self)

    # Extract RGB Bands
    symbology_x_batch = x_batch[:, symbology_bands]
    if stretch_type is not None:
        symbology_x_batch = image_batch_stretcher(
            symbology_x_batch, stretch_type, statistics_type
        )

    # Channel first to channel last and clamp float values to range 0 - 1 for plotting
    symbology_x_batch = symbology_x_batch.permute(0, 2, 3, 1)
    # Clamp float values to range 0 - 1
    if symbology_x_batch.mean() < 1:
        symbology_x_batch = symbology_x_batch.clamp(0, 1)

    # Squeeze channels if single channel (1, 224, 224) -> (224, 224)
    if symbology_x_batch.shape[-1] == 1:
        symbology_x_batch = symbology_x_batch.squeeze()

    # Get color Array
    color_array = self._multispectral_color_array

    # Size for plotting
    fig, axs = plt.subplots(
        nrows=nrows, ncols=ncols, figsize=(ncols * imsize, nrows * imsize)
    )
    idx = 0
    for r in range(nrows):
        for c in range(ncols):
            if idx < symbology_x_batch.shape[0]:
                axi = axs
                if nrows == 1:
                    axi = axi
                else:
                    axi = axi[r]
                if ncols == 1:
                    axi = axi
                else:
                    axi = axi[c]
                axi.imshow(symbology_x_batch[idx].cpu().numpy())

                if self.dataset_type == "MultiLabeled_Tiles":
                    one_hot_labels = y_batch[idx].tolist()
                    from itertools import compress

                    labels = compress(self.classes, one_hot_labels)
                    title = ";".join(labels)
                else:
                    title = f"{self.classes[y_batch[idx].item()]}"

                axi.set_title(title)
                axi.axis("off")
            else:
                axs[r][c].axis("off")
            idx += 1
    if is_arcgispronotebook():
        plt.show()


# Function to plot hard examples for multilabel classification
# This function has been taken from fastai and modified to work with multispectral and rgb (ArcGISMSImage).
def plot_multi_top_losses_modified(
    self, samples=3, figsize=(8, 8), save_misclassified=False
):
    "Show images in `top_losses` along with their prediction, actual, loss, and probability of predicted class in a multilabeled dataset."
    if samples > 20:
        print("Max 20 samples")
        return
    losses, idxs = self.top_losses(self.data.c)
    l_dim = len(losses.size())
    if l_dim == 1:
        losses, idxs = self.top_losses()
    (
        infolist,
        ordlosses_idxs,
        mismatches_idxs,
        mismatches,
        losses_mismatches,
        mismatchescontainer,
    ) = ([], [], [], [], [], [])
    truthlabels = np.asarray(self.y_true, dtype=int)
    classes_ids = [k for k in enumerate(self.data.classes)]
    predclass = np.asarray(self.pred_class)
    for i, pred in enumerate(predclass):
        where_truth = np.nonzero((truthlabels[i] > 0))[0]
        mismatch = np.all(pred != where_truth)
        if mismatch:
            mismatches_idxs.append(i)
            if l_dim > 1:
                losses_mismatches.append((losses[i][pred], i))
            else:
                losses_mismatches.append((losses[i], i))
        if l_dim > 1:
            infotup = (
                i,
                pred,
                where_truth,
                losses[i][pred],
                np.round(self.preds[i], decimals=3)[pred],
                mismatch,
            )
        else:
            infotup = (
                i,
                pred,
                where_truth,
                losses[i],
                np.round(self.preds[i], decimals=3)[pred],
                mismatch,
            )
        infolist.append(infotup)
    ds = self.data.dl(self.ds_type).dataset
    mismatches = ds[mismatches_idxs]
    ordlosses = sorted(losses_mismatches, key=lambda x: x[0], reverse=True)
    for w in ordlosses:
        ordlosses_idxs.append(w[1])
    mismatches_ordered_byloss = ds[ordlosses_idxs]
    print(
        f"{str(len(mismatches))} misclassified samples over {str(len(self.data.valid_ds))} samples in the validation set."
    )
    samples = min(samples, len(mismatches))
    from arcgis.learn._utils.common import ArcGISMSImage

    for ima in range(len(mismatches_ordered_byloss)):
        mismatchescontainer.append(mismatches_ordered_byloss[ima][0])
    for sampleN in range(samples):
        actualclasses = ""
        for clas in infolist[ordlosses_idxs[sampleN]][2]:
            actualclasses = f"{actualclasses} -- {str(classes_ids[clas][1])}"
        imag = mismatches_ordered_byloss[sampleN][0]
        imag = ArcGISMSImage.show(imag, return_ax=True)
        imag.set_title(
            f"""Predicted: {classes_ids[infolist[ordlosses_idxs[sampleN]][1]][1]} \nActual: {actualclasses}\nLoss: {infolist[ordlosses_idxs[sampleN]][3]}\nProbability: {infolist[ordlosses_idxs[sampleN]][4]}""",
            loc="left",
        )
        plt.show()
        if save_misclassified:
            return mismatchescontainer



from ._wms import WMSLayer
from .wmts import WMTSLayer
from ._csv import CSVLayer
from ._georss import GeoRSSLayer
from ._kml import KMLLayer
from ._geojson import GeoJSONLayer
from ._service import OGCCollection, OGCFeatureService

__all__ = [
    "WMTSLayer",
    "CSVLayer",
    "GeoRSSLayer",
    "KMLLayer",
    "WMSLayer",
    "GeoJSONLayer",
    "OGCCollection",
    "OGCFeatureService",
]



"""
IO operations for Feature Classes
"""
from arcgis.auth.tools import LazyLoader
import io
import os
import uuid
import copy
from pathlib import Path, PurePath
import logging
import datetime
import ujson as _ujson
import numpy as np
import pandas as pd
from contextlib import closing
import zipfile
from arcgis.geometry import Geometry

arcgis = LazyLoader("arcgis")
try:
    arcpy = LazyLoader("arcpy", strict=True)
    HASARCPY = True
except:
    HASARCPY = False

try:
    fiona = LazyLoader("fiona", strict=True)

    HASFIONA = True
except:
    HASFIONA = False

try:
    import shapefile

    HASPYSHP = True
    SHPVERSION = [int(i) for i in shapefile.__version__.split(".")]
except:
    HASPYSHP = False

_logging = logging.getLogger(__name__)


def _fc2pandas_dtypes(describe: dict) -> dict:
    """
    returns a dtypes to cast the final dataframe to the proper datatypes (if possible).

    :returns: Dict[str, Any]
    """
    if describe is None:
        return None
    if [float(i) for i in pd.__version__.split(".")] < [1, 0, 0]:
        _lu_types = {
            "OID": np.int64,
            "SmallInteger": np.int32,
            "Integer": np.int32,
            "Single": float,
            "Double": float,
            "String": "<U",
            "Blob": "O",
            "Guid": "<U38",
            "Raster": "O",
            "Date": "<M8[us]",
        }
    else:
        from numpy import dtype as _dtype

        _lu_types = {
            "OID": pd.Int64Dtype(),
            "SmallInteger": pd.Int32Dtype(),
            "Integer": pd.Int32Dtype(),
            "Single": float,
            "Double": float,
            "String": pd.StringDtype(),
            "Blob": "O",
            "Guid": pd.StringDtype(),
            "Raster": "O",
            "Date": "<M8[us]",
            "BigInteger": pd.Int64Dtype(),
            "DateOnly": _dtype("<M8[s]"),
            "TimeOnly": _dtype("<m8[s]"),
            "TimestampOffset": _dtype("<M8[us]"),
        }
    dtypes = None
    if "fields" in describe:
        dtypes = {}
        for field in describe["fields"]:
            # if field.type.lower() in ["TEXT", "string"]:
            #    dtypes[field.name] = f"{_lu_types['String']}{field.length}"
            if field.type in _lu_types.keys():
                dtypes[field.name] = _lu_types[field.type]
            elif field.type.lower() == "geometry":
                pass  # Skip value
            else:
                dtypes[field.name] = object
    return dtypes


# --------------------------------------------------------------------------
def _infer_type(df, col):
    """
    internal function used to get the datatypes for the feature class if
    the dataframe's _field_reference is NULL or there is a column that does
    not have a dtype assigned to it.

    Input:
     dataframe - Spatially Enabled DataFrame object
    Ouput:
      field type name
    """
    nn = df[col].notnull()
    nn = list(df[nn].index)
    if len(nn) > 0:
        val = df[col][nn[0]]
        if isinstance(val, str):
            return "TEXT"
        elif isinstance(val, tuple([int] + [np.int32])):
            return "INTEGER"
        elif isinstance(val, (float, np.int64)):
            return "FLOAT"
        elif isinstance(val, datetime):
            return "DATE"
    return "TEXT"


# --------------------------------------------------------------------------
def _geojson_to_esrijson(geojson):
    """converts the geojson spec to esri json spec"""
    if geojson["type"] in ["Polygon", "MultiPolygon"]:
        return {
            "rings": geojson["coordinates"],
            "spatialReference": {"wkid": 4326},
        }
    elif geojson["type"] == "Point":
        return {
            "x": geojson["coordinates"][0],
            "y": geojson["coordinates"][1],
            "spatialReference": {"wkid": 4326},
        }
    elif geojson["type"] == "MultiPoint":
        return {
            "points": geojson["coordinates"],
            "spatialReference": {"wkid": 4326},
        }
    elif geojson["type"] in ["LineString"]:  # , 'MultiLineString']:
        return {
            "paths": [[list(gj) for gj in geojson["coordinates"]]],
            "spatialReference": {"wkid": 4326},
        }
    elif geojson["type"] in ["MultiLineString"]:
        coords = []
        for pts in geojson["coordinates"]:
            coords.append(list([list(pt) for pt in pts]))
        return {"paths": coords, "spatialReference": {"wkid": 4326}}
    return geojson


# --------------------------------------------------------------------------
def _geometry_to_geojson(geom):
    """converts the esri json spec to geojson"""
    if "rings" in geom and len(geom["rings"]) == 1:
        return {"type": "Polygon", "coordinates": geom["rings"]}
    elif "rings" in geom and len(geom["rings"]) > 1:
        return {"type": "MultiPolygon", "coordinates": geom["rings"]}
    elif geom["type"] == "Point":
        return {"coordinates": [geom["x"], geom["y"]], "type": "Point"}
    elif geom["type"] == "MultiPoint":
        return {"coordinates": geom["points"], "type": "MultiPoint"}
    elif geom["type"].lower() == "polyline" and len(geom["paths"]) <= 1:
        return {"coordinates": geom["paths"], "type": "LineString"}
    elif geom["type"].lower() == "polyline" and len(geom["paths"]) > 1:
        return {"coordinates": geom["paths"], "type": "MultiLineString"}
    return geom


# --------------------------------------------------------------------------
def _from_xy(df, x_column, y_column, sr=None, z_column=None, m_column=None, **kwargs):
    """
    Takes an X/Y Column and Creates a Point Geometry from it. Can handle Z
    and M Columns as well.

    **kwargs**

    :oid_field: - Optional String value that can be specified to preven the int64 from turning into int32.


    """
    oid_field = kwargs.pop("oid_field", None)
    from arcgis.geometry import SpatialReference, Point
    from arcgis.features.geo._array import GeoArray

    def _xy_to_geometry(x, y, sr, z=None, m=None):
        """converts x/y coordinates to Point object"""
        if z and m:
            return Point({"spatialReference": sr, "x": x, "y": y, "z": z, "m": m})
        elif z and not m:
            return Point({"spatialReference": sr, "x": x, "y": y, "z": z})
        elif m and not z:
            return Point({"spatialReference": sr, "x": x, "y": y, "m": m})
        else:
            return Point({"spatialReference": sr, "x": x, "y": y})

    if sr is None:
        sr = SpatialReference({"wkid": 4326})
    if not isinstance(sr, SpatialReference):
        if isinstance(sr, dict):
            sr = SpatialReference(sr)
        elif isinstance(sr, int):
            sr = SpatialReference({"wkid": sr})
        elif isinstance(sr, str):
            sr = SpatialReference({"wkt": sr})
    geoms = []
    v_func = np.vectorize(_xy_to_geometry, otypes="O")
    ags_geom = np.empty(len(df), dtype="O")

    if z_column and m_column:
        ags_geom[:] = v_func(
            df[x_column].values,
            df[y_column].values,
            sr,
            df[z_column].values,
            df[m_column].values,
        )
    elif z_column and not m_column:
        ags_geom[:] = v_func(
            df[x_column].values,
            df[y_column].values,
            sr,
            df[z_column].values,
            None,
        )
    elif m_column and not z_column:
        ags_geom[:] = v_func(
            df[x_column].values,
            df[y_column].values,
            sr,
            None,
            df[m_column].values,
        )
    else:
        ags_geom[:] = v_func(df[x_column].values, df[y_column].values, sr)
    oid_fields = ["oid", "fid", "objectid"]
    if oid_field:
        oid_fields.append(oid_field.lower())
    df = df.astype(
        {
            col: "int32"
            for col in df.select_dtypes("int64").columns
            if not col.lower() in oid_fields
        }
    )
    df["SHAPE"] = GeoArray(ags_geom)
    df.spatial.name
    for i in range(len(df)):
        try:
            shape = df.iloc[i]["SHAPE"]
        except:
            shape = df.loc[i]["SHAPE"]
        if "EMPTY" in shape.WKT:
            df.iat[i, df.columns.get_loc("SHAPE")] = None
    return df


def _ensure_path_string(input_path):
    """Provide hander to facilitate file path inputs to be Path object instances."""
    return str(input_path) if isinstance(input_path, Path) else input_path


def from_url(url: str) -> list:
    """
    Loads a `shapefile` from a URL endpoint into a spatially enabled dataframe.

    ===========================     ====================================================================
    **Parameter**                    **Description**
    ---------------------------     --------------------------------------------------------------------
    url                             Required String. The web location of the compressed shapefile.
    ===========================     ====================================================================

    :return: List[pd.DataFrame]

    """

    if HASPYSHP == False:
        raise Exception("pyshp is required to read hosted shapefiles.")
    import requests

    r = requests.get(url)
    with closing(r), zipfile.ZipFile(io.BytesIO(r.content)) as archive:
        datasets = {}
        for member in archive.infolist():
            f_name_in_zip = member.filename
            filename = f_name_in_zip.split(".")[0]
            ends_with = os.path.splitext(member.filename)[1]
            if filename in datasets:
                if ends_with.lower() in [".shp", ".dbf", ".prj", ".shx"]:
                    datasets[filename][ends_with.replace(".", "")] = member
            else:
                datasets[filename] = {}
                if ends_with.lower() in [".shp", ".dbf", ".prj", ".shx"]:
                    datasets[filename][ends_with.replace(".", "")] = member
        # build readers
        readers = []
        for key in datasets.keys():
            if list(datasets[key].keys()) >= ["shp", "dbf"]:
                shx = None
                if datasets[key].get("shx", None):
                    shx = io.BytesIO()
                    shx.write(archive.read(datasets[key].get("shx", None).filename))

                shp = io.BytesIO()  #
                shp.write(archive.read(datasets[key].get("shp", None).filename))
                dbf = io.BytesIO()
                dbf.write(archive.read(datasets[key].get("dbf", None).filename))
                readers.append(shapefile.Reader(shp=shp, shx=shx, dbf=dbf))
        # construct SeDF from URL based datasets
        sdfs = []
        for reader in readers:
            records = []
            fields = [field[0] for field in reader.fields if field[0] != "DeletionFlag"]
            for idx, r in enumerate(reader.shapeRecords()):
                atr = dict(zip(fields, r.record))
                g = r.shape.__geo_interface__
                geom = Geometry(g)
                atr["SHAPE"] = geom
                records.append(atr)
                del atr
                del r, g
                del geom
            sdf = pd.DataFrame(records)
            sdf.spatial.set_geometry("SHAPE")
            sdf["OBJECTID"] = range(sdf.shape[0])
            sdf.reset_index(inplace=True)
            sdf.spatial._meta.source = url
            sdfs.append(sdf)
            del reader
        return sdfs


# --------------------------------------------------------------------------
def read_feather(
    path, spatial_column="SHAPE", columns=None, use_threads: bool = True
) -> pd.DataFrame:
    """
    Load a feather-format object from the file path.

    Parameters
    ----------
    path : str, path object or file-like object
        Any valid string path is acceptable. The string could be a URL. Valid
        URL schemes include http, ftp, s3, and file. For file URLs, a host is
        expected. A local file could be:
        ``file://localhost/path/to/table.feather``.

        If you want to pass in a path object, pandas accepts any
        ``os.PathLike``.

        By file-like object, we refer to objects with a ``read()`` method,
        such as a file handler (e.g. via builtin ``open`` function)
        or ``StringIO``.
    spatial_column : str, Name of the geospatial column. The default is `SHAPE`.
       .. versionadded:: v1.8.2 of ArcGIS API for Python
    columns : sequence, default None
        If not provided, all columns are read.

        .. versionadded:: v1.8.2 of ArcGIS API for Python
    use_threads : bool, default True
        Whether to parallelize reading using multiple threads.

       .. versionadded:: v1.8.2 of ArcGIS API for Python

    Returns
    -------
    type of object stored in file
    """
    sdf = pd.read_feather(path=path, columns=columns, use_threads=use_threads)
    if spatial_column and spatial_column in sdf.columns:
        sdf.spatial.set_geometry(spatial_column)
        sdf.spatial.name
    return sdf


# --------------------------------------------------------------------------
def from_table(filename, **kwargs):
    """
    Allows a user to read from a non-spatial table

    **Note: ArcPy is Required for this method**

    ===============     ====================================================
    **Parameter**        **Description**
    ---------------     ----------------------------------------------------
    filename            Required string or pathlib.Path. The path to the
                        table.
    ===============     ====================================================

    **Keyword Arguments**

    ===============     ====================================================
    **Parameter**        **Description**
    ---------------     ----------------------------------------------------
    fields              Optional List/Tuple. A list (or tuple) of field
                        names. For a single field, you can use a string
                        instead of a list of strings.

                        Use an asterisk (*) instead of a list of fields if
                        you want to access all fields from the input table
                        (raster and BLOB fields are excluded). However, for
                        faster performance and reliable field order, it is
                        recommended that the list of fields be narrowed to
                        only those that are actually needed.

                        Geometry, raster, and BLOB fields are not supported.

    ---------------     ----------------------------------------------------
    where               Optional String. An optional expression that limits
                        the records returned.
    ---------------     ----------------------------------------------------
    skip_nulls          Optional Boolean. This controls whether records
                        using nulls are skipped. Default is True.
    ---------------     ----------------------------------------------------
    null_value          Optional String/Integer/Float. Replaces null values
                        from the input with a new value.
    ===============     ====================================================

    :return: pd.DataFrame

    """

    filename = _ensure_path_string(filename)

    if HASARCPY and not filename.lower().endswith(".dbf"):
        import arcpy

        where = kwargs.pop("where", None)
        fields = kwargs.pop("fields", "*")
        skip_nulls = kwargs.pop("skip_nulls", True)
        null_value = kwargs.pop("null_value", None)
        arr = arcpy.da.TableToNumPyArray(
            in_table=filename,
            field_names=fields,
            where_clause=where,
            skip_nulls=skip_nulls,
            null_value=null_value,
        )
        return pd.DataFrame(arr)
    elif HASARCPY and filename.lower().endswith(".dbf"):
        import arcpy

        with arcpy.da.SearchCursor(
            in_table=filename,
            field_names=kwargs.pop("fields", "*"),
            where_clause=kwargs.pop("where", None),
        ) as scur:
            array = [row for row in scur]
            df = pd.DataFrame(array, columns=scur.fields)
            try:
                return df.convert_dtypes()
            except:
                return df
        return None
    elif filename.lower().endswith(".dbf"):
        import shapefile

        with open(filename, "rb") as f:
            reader = shapefile.Reader(dbf=f)
            return pd.DataFrame([record.as_dict() for record in reader.iterRecords()])
    elif filename.lower().find(".csv") > -1:
        return pd.read_csv(filename)

    return


# --------------------------------------------------------------------------
def to_table(geo, location, overwrite=True, sanitize_columns=False):
    """
    Exports a geo enabled dataframe to a table.

    .. note::
        Null integer values will be changed to 0 when using shapely instead
        of ArcPy due to shapely conventions.
        With ArcPy null integer values will remain null.

    ===========================     ====================================================================
    **Parameter**                    **Description**
    ---------------------------     --------------------------------------------------------------------
    location                        Required string. The output of the table.
    ---------------------------     --------------------------------------------------------------------
    overwrite                       Optional Boolean.  If True and if the table exists, it will be
                                    deleted and overwritten.  This is default.  If False, the table and
                                    the table exists, and exception will be raised.
    ---------------------------     --------------------------------------------------------------------
    sanitize_columns                Optional Boolean. If True, column names will be converted to
                                    string, invalid characters removed and other checks will be
                                    performed. The default is False.
    ===========================     ====================================================================

    :return: String
    """
    old_column, old_index = None, None
    if sanitize_columns:
        old_column = geo._data.columns.tolist()
        old_index = copy.deepcopy(geo._data.index)
        _sanitize_column_names(geo, inplace=True)
    out_location = os.path.dirname(location)
    fc_name = os.path.basename(location)
    df = geo._data.copy().convert_dtypes()
    df[df.select_dtypes(np.number).columns.tolist()] = df[
        df.select_dtypes(np.number).columns.tolist()
    ].replace({pd.NA: None})
    df[df.select_dtypes(pd.StringDtype()).columns.tolist()] = df[
        df.select_dtypes(pd.StringDtype()).columns.tolist()
    ].replace(pd.NA, "")

    if location.lower().find(".csv") > -1:
        geo._data.to_csv(location)
        if not old_column is None:
            geo._data.columns = old_column
        if not old_index is None:
            geo._data.index = old_index
        return location
    elif HASARCPY:
        import arcpy

        columns = df.columns.tolist()
        join_dummy = "AEIOUYAJC81Z"
        try:
            columns.pop(columns.index(df.spatial.name))
        except:
            pass
        dtypes = [(join_dummy, np.int64)]
        if overwrite and arcpy.Exists(location):
            arcpy.Delete_management(location)
        elif overwrite == False and arcpy.Exists(location):
            raise ValueError(
                ("overwrite set to False, Cannot " "overwrite the table. ")
            )
        fc = arcpy.CreateTable_management(out_path=out_location, out_name=fc_name)[0]
        # 2. Add the Fields and Data Types
        #
        oidfld = arcpy.da.Describe(fc)["OIDFieldName"]
        for col in columns[:]:
            if (col.lower() == oidfld.lower()) or (
                col.lower() in ["fid", "oid", "objectid"] and location.endswith(".dbf")
            ):
                pass
            elif col.lower() in ["fid", "oid", "objectid"]:
                dtypes.append((col, np.int32))
            elif df[col].dtype.name == "datetime64[s]":
                dtypes.append((col, "<m8[us]"))
            elif df[col].dtype.name.find("datetime") > -1:
                dtypes.append((col, "<M8[us]"))
                df[col] = df[col].dt.to_pydatetime()
            elif df[col].dtype.name.find("timedelta") > -1:
                dtypes.append((col, float))
                df[col] = df[col].dt.total_seconds() * 1000
            elif df[col].dtype.name == "object":
                try:
                    u = type(df[col][df[col].first_valid_index()])
                except:
                    u = pd.unique(df[col].apply(type)).tolist()[0]
                if u is None:
                    dtypes.append((col, "<U254"))
                elif u == datetime.time:
                    dtypes.append((col, "<m8[us]"))
                elif u == type(None):
                    dtypes.append((col, "<U254"))
                elif issubclass(u, str):
                    mlen = df[col].str.len().max()
                    dtypes.append((col, "<U%s" % int(mlen)))
                else:
                    try:
                        dtypes.append((col, type(df[col][0])))
                    except:
                        dtypes.append((col, "<U254"))
            elif df[col].dtype.name == "string":
                try:
                    u = type(df[col][df[col].first_valid_index()])
                except:
                    u = pd.unique(df[col].apply(type)).tolist()[0]
                if u is None:
                    dtypes.append((col, "<U254"))
                elif u == datetime.time:
                    dtypes.append((col, "<m8[us]"))
                elif issubclass(u, str):
                    mlen = df[col].str.len().max()
                    if int(mlen) == 0:
                        mlen = 1
                    dtypes.append((col, "<U%s" % int(mlen)))
                else:
                    try:
                        dtypes.append((col, type(df[col][0])))
                    except:
                        dtypes.append((col, "<U254"))
            elif df[col].dtype.name in ["int64", "Int64"]:
                dtypes.append((col, "<i8"))
            elif df[col].dtype.name == "bool":
                dtypes.append((col, np.int32))
            elif df[col].dtype.name == "boolean":
                dtypes.append((col, np.int32))
            elif isinstance(df[col].dtype, pd.CategoricalDtype):
                dtype = df[col].dtype
                if dtype.categories.dtype.name == "object":
                    try:
                        msize = max(dtype.categories.str.len())
                    except:
                        msize = 254
                    dtypes.append((col, "<U%s" % msize))
                elif dtype.categories.dtype.name.find("datetime") > -1:
                    dtypes.append((col, "<M8[us]"))
                else:
                    dtypes.append((col, dtype.categories.dtype))
            else:
                dtypes.append((col, df[col].dtype.type))

        array = np.array([], np.dtype(dtypes))
        arcpy.da.ExtendTable(fc, oidfld, array, join_dummy, append_only=False)
        # 3. Insert the Data
        #
        fields = arcpy.ListFields(fc)
        icols = [
            fld.name
            for fld in fields
            if fld.type not in ["OID", "Geometry", "FID"] and fld.name in df.columns
        ]
        dfcols = [
            fld.name
            for fld in fields
            if fld.type not in ["OID", "Geometry", "FID"] and fld.name in df.columns
        ]
        with arcpy.da.InsertCursor(fc, icols) as irows:
            bool_fld_idx = [
                irows.fields.index(col)
                for col in df.select_dtypes(pd.BooleanDtype()).columns.tolist()
            ]
            dt_fld_idx = [
                irows.fields.index(col)
                for col in df.columns
                if df[col].dtype.name.startswith("datetime")
            ]
            if len(bool_fld_idx) > 0:
                df = df.replace({pd.NA: None})
            if len(dt_fld_idx) > 0:
                df = df.replace({pd.NaT: None})
            for idx, row in df[dfcols].iterrows():
                try:
                    row = row.tolist()
                    if len(dt_fld_idx) > 0:
                        for idx in dt_fld_idx:
                            if row[idx]:
                                row[idx] = row[idx].to_pydatetime()
                    if len(bool_fld_idx) > 0:
                        for idx in bool_fld_idx:
                            if row[idx]:
                                row[idx] = int(row[idx])
                    irows.insertRow(row)

                except:
                    _logging.warn("row %s could not be inserted." % idx)
        if not old_column is None:
            geo._data.columns = old_column
        if not old_index is None:
            geo._data.index = old_index
        return fc
    if not old_column is None:
        geo._data.columns = old_column
    if not old_index is None:
        geo._data.index = old_index
    return


# --------------------------------------------------------------------------
def from_featureclass(filename, **kwargs):
    """
    Returns a GeoDataFrame (Spatially Enabled Pandas DataFrame) from a feature class.

    ===========================     ====================================================================
    **Parameter**                    **Description**
    ---------------------------     --------------------------------------------------------------------
    filename                        Required string or pathlib.Path. Full path to the feature class or URL (shapefiles only).
    ===========================     ====================================================================

    *Optional parameters when ArcPy library is available in the current environment*:
    ===========================     ====================================================================
    **Key**                         **Value**
    ---------------------------     --------------------------------------------------------------------
    sql_clause                      sql clause to parse data down. To learn more see
                                    [ArcPy Search Cursor](https://pro.arcgis.com/en/pro-app/arcpy/data-access/searchcursor-class.htm)
    ---------------------------     --------------------------------------------------------------------
    where_clause                    where statement. To learn more see [ArcPy SQL reference](https://pro.arcgis.com/en/pro-app/help/mapping/navigation/sql-reference-for-elements-used-in-query-expressions.htm)
    ---------------------------     --------------------------------------------------------------------
    fields                          list of strings specifying the field names.
    ---------------------------     --------------------------------------------------------------------
    spatial_filter                  A `Geometry` object that will filter the results.  This requires
                                    `arcpy` to work.
    ---------------------------     --------------------------------------------------------------------
    sr                              A Spatial reference to project (or tranform) output GeoDataFrame to.
                                    This requires `arcpy` to work.
    ---------------------------     --------------------------------------------------------------------
    datum_transformation            Used in combination with 'sr' parameter. if the spatial reference of
                                    output GeoDataFrame and input data do not share the same datum,
                                    an appropriate datum transformation should be specified.
                                    To Learn more see [Geographic datum transformations](https://pro.arcgis.com/en/pro-app/help/mapping/properties/geographic-coordinate-system-transformation.htm)
                                    This requires `arcpy` to work.
    ===========================     ====================================================================

    :return: pandas.core.frame.DataFrame

    """
    from arcgis.geometry import _types
    import json

    if HASARCPY and (
        isinstance(filename, (arcpy._mp.Layer))
        or type(filename).__name__.find("arcpy") > -1
    ):
        filename = filename
    else:
        filename = _ensure_path_string(filename)
        if not isinstance(filename, (str, Path, PurePath)):
            raise ValueError(
                f"filename must be a `str`, `Path`, or `PurePath`, not {type(filename)}"
            )
        if filename.find("http://") > -1 or filename.find("https://") > -1:
            res = from_url(url=filename)
            if len(res) == 1:
                return res[0]
            return res
    if HASARCPY:
        sql_clause = kwargs.pop("sql_clause", (None, None))
        where_clause = kwargs.pop("where_clause", None)
        fields = kwargs.pop("fields", None)
        sr = kwargs.pop("sr", None)
        spatial_filter = kwargs.pop("spatial_filter", None)
        geom = None
        try:
            desc = arcpy.da.Describe(filename)
            area_field = desc.pop("areaFieldName", None)
            length_field = desc.pop("lengthFieldName", None)
        except:  # for older versions of arcpy
            desc = arcpy.Describe(filename)
            desc = {"fields": desc.fields, "shapeType": desc.shapeType}
            area_field = getattr(desc, "areaFieldName", None)
            length_field = getattr(desc, "lengthFieldName", None)
        pandas_dtypes = _fc2pandas_dtypes(desc)

        if spatial_filter:
            _sf_lu = {
                "esriSpatialRelIntersects": "INTERSECT",
                "esriSpatialRelContains": "CONTAINS",
                "esriSpatialRelCrosses": "CROSSED_BY_THE_OUTLINE_OF",
                "esriSpatialRelEnvelopeIntersects": "INTERSECT",
                "esriSpatialRelIndexIntersects": "INTERSECT",
                "esriSpatialRelOverlaps": "INTERSECT",
                "esriSpatialRelTouches": "BOUNDARY_TOUCHES",
                "esriSpatialRelWithin": "WITHIN",
            }
            relto = _sf_lu[spatial_filter["spatialRel"]]
            geom = spatial_filter["geometry"]
            if hasattr(geom, "polygon"):
                geom = geom.polygon
            geom = geom.as_arcpy
            flname = "a" + uuid.uuid4().hex[:6]
            filename = arcpy.management.MakeFeatureLayer(
                filename, out_layer=flname, where_clause=where_clause
            )[0]
            arcpy.management.SelectLayerByLocation(
                filename, overlap_type=relto, select_features=geom
            )[0]

        shape_name = desc["shapeType"]
        if fields is None:
            fields = [
                fld.name
                for fld in desc["fields"]
                if fld.type not in ["Geometry"]
                and fld.name not in [area_field, length_field]
            ]
        cursor_fields = fields + ["SHAPE@JSON"]
        df_fields = fields + ["SHAPE"]
        count = 0
        dfs = []
        shape_field_idx = cursor_fields.index("SHAPE@JSON")
        with arcpy.da.SearchCursor(
            filename,
            field_names=cursor_fields,
            where_clause=where_clause,
            sql_clause=sql_clause,
            spatial_reference=sr,
            datum_transformation=kwargs.get("datum_transformation", None),
        ) as rows:
            srows = []
            for row in rows:
                srows.append(row)
                if len(srows) == 25000:
                    dfs.append(pd.DataFrame(srows, columns=df_fields))
                    srows = []
            if len(srows):
                dfs.append(pd.DataFrame(srows, columns=df_fields))
                srows = []
            del srows
        if len(dfs) > 0:
            df = pd.concat(dfs)
            df = df.reset_index(drop=True)
        elif len(dfs) == 1:
            df = dfs[0]
        else:
            df = pd.DataFrame([], columns=df_fields)
        q = df.SHAPE.notnull()
        none_q = ~df.SHAPE.notnull()
        gt = desc["shapeType"].lower()
        geoms = {
            "point": _types.Point,
            "polygon": _types.Polygon,
            "polyline": _types.Polyline,
            "multipoint": _types.MultiPoint,
            "envelope": _types.Envelope,
            "geometry": _types.Geometry,
        }
        import json

        df.SHAPE = df.SHAPE[q].apply(_ujson.loads).apply(geoms[gt])
        df.loc[none_q, "SHAPE"] = None
        df.spatial.set_geometry("SHAPE")
        df.spatial._meta.source = filename
        try:
            for key, value in pandas_dtypes.items():
                try:
                    df[key] = df[key].astype(value)  # , errors='ignore')
                except:
                    from numpy import dtype as _dtype

                    if value == _dtype("<m8[s]"):
                        df[key] = pd.to_datetime(df[key], format="%H:%M:%S").dt.time
                    elif value == _dtype("<M8[us]"):
                        df[key] = pd.to_datetime(df[key], utc=True)

            return df.convert_dtypes()
        except:
            return df.convert_dtypes()
    elif HASARCPY == False and HASPYSHP == True and filename.lower().find(".shp") > -1:
        geoms = []
        records = []
        reader = shapefile.Reader(filename)
        fields = [field[0] for field in reader.fields if field[0] != "DeletionFlag"]
        for r in reader.shapeRecords():
            atr = dict(zip(fields, r.record))
            g = r.shape.__geo_interface__
            g = _geojson_to_esrijson(g)
            geom = _types.Geometry(g)
            atr["SHAPE"] = geom
            records.append(atr)
            del atr
            del r, g
            del geom
        sdf = pd.DataFrame(records)
        sdf.spatial.set_geometry("SHAPE")
        sdf["OBJECTID"] = range(sdf.shape[0])
        sdf.reset_index(inplace=True)
        sdf.spatial._meta.source = filename
        return sdf
    elif (
        HASARCPY == False
        and HASFIONA == True
        and (
            filename.lower().find(".shp") > -1
            or os.path.dirname(filename).lower().find(".gdb") > -1
        )
    ):
        is_gdb = os.path.dirname(filename).lower().find(".gdb") > -1
        if is_gdb:
            # Remove deprecation warning.
            fiona_env = fiona.drivers
            if hasattr(fiona, "Env"):
                fiona_env = fiona.Env

            with fiona_env():
                from arcgis.geometry import _types

                fp = os.path.dirname(filename)
                fn = os.path.basename(filename)
                geoms = []
                atts = []
                with fiona.open(fp, layer=fn) as source:
                    meta = source.meta
                    cols = list(source.schema["properties"].keys())

                    # Get the CRS
                    try:
                        wkid = source.crs["init"].split(":")[1]
                    except:
                        wkid = 4326

                    sr = _types.SpatialReference({"wkid": int(wkid)})

                    for idx, row in source.items():
                        g = _types.Geometry(row["geometry"])
                        geoms.append(g)
                        atts.append(list(row["properties"].values()))
                        del idx, row
                    df = pd.DataFrame(data=atts, columns=cols)
                    df.spatial.set_geometry(geoms)
                    df.spatial.sr = sr
                    df.spatial._meta.source = filename
                    return df
        else:
            with fiona.drivers():
                from arcgis.geometry import _types

                geoms = []
                atts = []
                with fiona.open(filename) as source:
                    meta = source.meta
                    cols = list(source.schema["properties"].keys())
                    for idx, row in source.items():
                        geoms.append(_types.Geometry(row["geometry"]))
                        atts.append(list(row["properties"].values()))
                        del idx, row
                    df = pd.DataFrame(data=atts, columns=cols)
                    df.spatial.set_geometry(geoms)
                    df.spatial._meta.source = filename
                    return df
    else:
        if os.path.dirname(filename).lower().find(".gdb") > -1:
            message = """
            Cannot Open Geodatabase without Arcpy or Fiona
            \nPlease switch to Arcpy for full support or install fiona by this command `conda install fiona`
            """.strip()
            print(message)
            raise Exception("Failed to import Feature Class from Geodatabase specified")
        else:
            raise Exception(
                "Unsupported Data Format or Invalid Feature Class specified"
            )
    # return


# --------------------------------------------------------------------------
import functools


@functools.lru_cache(maxsize=255)
def _examine_meta(meta):
    gt_lu: dict[str, str] = {
        "esriGeometryPoint": "point",
        "esriGeometryLine": "polyline",
        "esriGeometryPolyline": "polyline",
        "esriGeometryPath": "polyline",
        "esriGeometryPolygon": "polygon",
        "esriGeometryMultiPatch": "polygon",
        "esriGeometryMultipoint": "multipoint",
    }
    sr = arcgis.geometry.SpatialReference({"wkid": 4326})
    if isinstance(meta.source, arcgis.features.FeatureLayer):
        try:
            sr = arcgis.geometry.SpatialReference(
                meta.source.properties["extent"]["spatialReference"]
            )
        except:
            sr = arcgis.geometry.SpatialReference({"wkid": 4326})

        return (gt_lu[meta.source.properties.geometryType], sr)
    return "point", sr


# --------------------------------------------------------------------------
def to_featureclass(
    geo,
    location,
    overwrite=True,
    validate=False,
    sanitize_columns=True,
    has_m=True,
    has_z=False,
):
    """
    Exports the DataFrame to a Feature class.

    .. note::
        Null integer values will be changed to 0 when using shapely instead
        of ArcPy due to shapely conventions.
        With ArcPy null integer values will remain null.

    ===============     ====================================================
    **Parameter**        **Description**
    ---------------     ----------------------------------------------------
    location            Required string. This is the output location for the
                        feature class. This should be the path and feature
                        class name.
    ---------------     ----------------------------------------------------
    overwrite           Optional Boolean. If overwrite is true, existing
                        data will be deleted and replaced with the spatial
                        dataframe.
    ---------------     ----------------------------------------------------
    validate            Optional Boolean. If True, the export will check if
                        all the geometry objects are correct upon export.
    ---------------     ----------------------------------------------------
    sanitize_columns    Optional Boolean. If True, column names will be
                        converted to string, invalid characters removed and
                        other checks will be performed. The default is False.
    ---------------     ----------------------------------------------------
    ham_m               Optional Boolean to indicate if data has linear
                        referencing (m) values. Default is False.
    ---------------     ----------------------------------------------------
    has_z               Optional Boolean to indicate if data has elevation
                        (z) values. Default is False.
    ===============     ====================================================


    :return: A string

    """
    out_location = os.path.dirname(location)

    fc_name = os.path.basename(location)
    df = geo._data.copy().convert_dtypes()
    old_idx = df.index
    df.reset_index(drop=True, inplace=True)
    if geo.name is None:
        raise ValueError("DataFrame must have geometry set.")
    if validate and geo.validate(strict=True) == False:
        raise ValueError(
            ("Mixed geometry types detected, " "cannot export to feature class.")
        )
    # deep copy of original columns to reassign them in finally of arcpy statement
    original_columns = copy.deepcopy(df.columns.tolist())
    geometry_name = geo.name or None
    if geometry_name and geometry_name != "SHAPE":
        new_columns = geo._data.columns.tolist()
        new_columns[new_columns.index(geometry_name)] = "SHAPE"
        df.columns = new_columns
        df.spatial._name = "SHAPE"
    # sanitize
    if sanitize_columns:
        # logic
        _sanitize_column_names(df.spatial, inplace=True)

    columns = df.columns.tolist()
    for col in columns[:]:
        if not isinstance(col, str):
            df.rename(columns={col: str(col)}, inplace=True)
            col = str(col)
    df[df.select_dtypes(np.number).columns.tolist()] = df[
        df.select_dtypes(np.number).columns.tolist()
    ].replace({pd.NA: None})
    df[df.select_dtypes(pd.StringDtype()).columns.tolist()] = df[
        df.select_dtypes(pd.StringDtype()).columns.tolist()
    ].replace(pd.NA, "")

    if HASARCPY:
        try:
            # 1. Create the Save Feature Class
            #
            columns = df.columns.tolist()
            join_dummy = "AEIOUYAJC81Z"
            columns.pop(columns.index(df.spatial.name))
            dtypes = [(join_dummy, np.int64)]
            if overwrite and arcpy.Exists(location):
                arcpy.Delete_management(location)
            elif overwrite == False and arcpy.Exists(location):
                raise ValueError(
                    ("overwrite set to False, Cannot " "overwrite the table. ")
                )

            notnull = df[df.spatial.name].notnull()
            idx = df[df.spatial.name][notnull].first_valid_index()
            if idx is None:
                gt, sr = _examine_meta(df.spatial._meta)
            else:
                sr = arcgis.geometry.SpatialReference(
                    df[df.spatial.name][idx]["spatialReference"]
                )
                gt = df[df.spatial.name][idx].geometry_type.upper()

            null_geom = {
                "point": pd.io.json.dumps(
                    {"x": None, "y": None, "spatialReference": sr}
                ),
                "polyline": pd.io.json.dumps({"paths": [], "spatialReference": sr}),
                "polygon": pd.io.json.dumps({"rings": [], "spatialReference": sr}),
                "multipoint": pd.io.json.dumps({"points": [], "spatialReference": sr}),
            }

            null_geom = null_geom[gt.lower()]
            sr = sr.as_arcpy
            if has_m == True:
                has_m = "ENABLED"
            else:
                has_m = None

            if has_z == True:
                has_z = "ENABLED"
            else:
                has_z = None

            fc = arcpy.CreateFeatureclass_management(
                out_location,
                spatial_reference=sr,
                geometry_type=gt,
                out_name=fc_name,
                has_m=has_m,
                has_z=has_z,
            )[0]

            # 2. Add the Fields and Data Types
            oidfld = arcpy.da.Describe(fc)["OIDFieldName"]
            for col in columns[:]:
                if col.lower() in ["fid", "oid", "objectid"]:
                    dtypes.append((col, np.int32))
                elif df[col].dtype.name == "datetime64[s]":
                    dtypes.append((col, "<m8[us]"))
                elif df[col].dtype.name.startswith("datetime"):
                    dtypes.append((col, "<M8[us]"))
                elif df[col].dtype.name.find("timedelta") > -1:
                    dtypes.append((col, float))
                    df[col] = df[col].dt.total_seconds() * 1000
                elif df[col].dtype.name == "object":
                    try:
                        u = type(df[col][df[col].first_valid_index()])
                    except:
                        u = pd.unique(df[col].apply(type)).tolist()[0]
                    if u is None:
                        dtypes.append((col, "<U254"))
                    elif u == datetime.time:
                        dtypes.append((col, "<m8[us]"))
                    elif issubclass(u, str):
                        mlen = df[col].str.len().max()
                        dtypes.append((col, "<U%s" % int(mlen)))
                    elif u is datetime.datetime:
                        dtypes.append((col, "<M8[us]"))
                    else:
                        try:
                            if df[col][idx] is None:
                                dtypes.append((col, "<U254"))
                            else:
                                dtypes.append((col, type(df[col][idx])))
                        except:
                            dtypes.append((col, "<U254"))
                elif df[col].dtype.name in ["int64", "Int64"]:
                    # Enterprise 11.1 and less do not accept Int64. Need to make float
                    gis = arcgis.env.active_gis
                    if (
                        gis is not None
                        and gis._is_agol == False
                        and gis.version <= [10, 3]
                    ):
                        dtypes.append((col, np.float64))
                    else:
                        dtypes.append((col, "<i8"))
                elif df[col].dtype.name == "bool":
                    dtypes.append((col, np.int32))
                elif df[col].dtype.name == "boolean":
                    dtypes.append((col, np.int32))
                elif isinstance(df[col].dtype, pd.CategoricalDtype):
                    dtype = df[col].dtype
                    if dtype.categories.dtype.name == "object":
                        try:
                            msize = max(dtype.categories.str.len())
                        except:
                            msize = 254
                        dtypes.append((col, "<U%s" % msize))
                    elif dtype.categories.dtype.name.find("datetime") > -1:
                        dtypes.append((col, "<M8[us]"))
                    else:
                        dtypes.append((col, dtype.categories.dtype))
                else:
                    if (
                        df[col].dtype.name == "object"
                        and df[col].first_valid_index()
                        and isinstance(df[col][idx], datetime.datetime)
                    ):
                        dtypes.append((col, "<M8[us]"))
                    elif df[col].dtype.type == str:
                        mlen = df[col].str.len().max()
                        if pd.isna(mlen):
                            mlen = 254
                        elif mlen == 0:
                            mlen = 254
                        dtypes.append((col, "<U%s" % int(mlen)))
                    else:
                        dtypes.append((col, df[col].dtype.type))
            from arcgis._impl.common._utils import chunks as _chunks

            smaller_dtypes = [[dtypes[0]] + flds for flds in _chunks(dtypes[1:], 10)]
            smaller_array = [np.array([], np.dtype(d)) for d in smaller_dtypes]
            for array in smaller_array:
                try:
                    arcpy.da.ExtendTable(
                        fc, oidfld, array, join_dummy, append_only=False
                    )
                except Exception as e:
                    print(e)

            # 3. Insert the Data
            if len(df) == 0:
                return fc
            fields = arcpy.ListFields(fc)
            icols = [
                fld.name
                for fld in fields
                if fld.type not in ["OID", "Geometry"] and fld.name in df.columns
            ] + ["SHAPE@JSON"]
            dfcols = [
                fld.name
                for fld in fields
                if fld.type not in ["OID", "Geometry"] and fld.name in df.columns
            ] + [df.spatial.name]

            with arcpy.da.InsertCursor(fc, icols) as irows:
                dt_fld_idx = [
                    irows.fields.index(col)
                    for col in df.columns
                    if df[col].dtype.name.startswith("datetime")
                ]
                bool_fld_idx = [
                    irows.fields.index(col)
                    for col in df.select_dtypes(pd.BooleanDtype()).columns.tolist()
                ]
                if len(bool_fld_idx) > 0:
                    df = df.replace({pd.NA: None})
                if len(dt_fld_idx) > 0:
                    df = df.replace({pd.NaT: None})

                def _insert_row(row):
                    row[-1] = pd.io.json.dumps(row[-1])
                    for idx in bool_fld_idx:
                        if isinstance(row[idx], (int, bool)):
                            row[idx] = int(row[idx])
                        else:
                            row[idx] = None
                    try:
                        irows.insertRow(row)
                    except Exception as e:
                        _logging.warn(
                            f"Could not insert the row because of error message: {e}. Recheck your data."
                        )

                q = df[df.spatial.name].isna()
                df.loc[q, "SHAPE"] = null_geom  # set null values to proper JSON
                replace_mappings = {
                    pd.NA: None,
                    np.nan: None,
                    np.NaN: None,
                    np.NAN: None,
                    pd.NaT: None,
                }
                np.apply_along_axis(
                    _insert_row,
                    1,
                    df.replace(replace_mappings)[dfcols].values,
                )

                df.loc[q, "SHAPE"] = None  # reset null values
        except ValueError as ve:
            df.columns = original_columns
            df.set_index(old_idx)
            fc = None
            raise
        except Exception as e:
            # something failed in try so reset columns to original columns
            # return empty item
            fc = None
            raise e
        finally:
            df.columns = original_columns
            df.set_index(old_idx)
        return fc
    elif HASPYSHP:
        if fc_name.endswith(".shp") == False:
            fc_name = "%s.shp" % fc_name
        if SHPVERSION < [2]:
            res = _pyshp_to_shapefile(df=df, out_path=out_location, out_name=fc_name)
            df.set_index(old_idx)
            return res
        else:
            res = _pyshp2(df=df, out_path=out_location, out_name=fc_name)
            df.set_index(old_idx)
            return res
    elif HASARCPY == False and HASPYSHP == False:
        raise Exception(
            (
                "Cannot Export the data without ArcPy or PyShp modules."
                " Please install them and try again."
            )
        )
    else:
        df.set_index(old_idx)
        return None


# --------------------------------------------------------------------------
def _pyshp_to_shapefile(df, out_path, out_name):
    """
    Saves a Spatially Enabled DataFrame to a Shapefile using pyshp

    :Parameters:
     :df: spatial dataframe
     :out_path: folder location to save the data
     :out_name: name of the shapefile
    :Output:
     path to the shapefile or None if pyshp isn't installed or
     spatial dataframe does not have a geometry column.
    """
    from arcgis.geometry._types import Geometry

    if HASPYSHP:
        GEOMTYPELOOKUP = {
            "Polygon": shapefile.POLYGON,
            "Point": shapefile.POINT,
            "Polyline": shapefile.POLYLINE,
            "null": shapefile.NULL,
        }
        if os.path.isdir(out_path) == False:
            os.makedirs(out_path)
        out_fc = os.path.join(out_path, out_name)
        if out_fc.lower().endswith(".shp") == False:
            out_fc += ".shp"
        geom_field = df.spatial.name
        if geom_field is None:
            return
        geom_type = "null"
        idx = df[geom_field].first_valid_index()
        if idx > -1:
            geom_type = df.loc[idx][geom_field].type
        with shapefile.Writer(
            GEOMTYPELOOKUP[geom_type], dbf=out_fc + ".dbf"
        ) as shpfile:
            shpfile.autoBalance = 1
            dfields = []
            cfields = []
            for c in df.columns:
                idx = df[c].first_valid_index() or df.index.tolist()[0]
                if idx > -1:
                    if isinstance(df[c].loc[idx], Geometry):
                        geom_field = (c, "GEOMETRY")
                    else:
                        cfields.append(c)
                        if isinstance(df[c].loc[idx], (str)):
                            shpfile.field(name=c, size=255)
                        elif isinstance(df[c].loc[idx], (int)):
                            shpfile.field(name=c, fieldType="N", size=5)
                        elif isinstance(df[c].loc[idx], (int, np.int32)):
                            shpfile.field(name=c, fieldType="N", size=10)
                        elif isinstance(df[c].loc[idx], (float, np.float64, np.int64)):
                            shpfile.field(name=c, fieldType="F", size=19, decimal=11)
                        elif (
                            isinstance(
                                df[c].loc[idx],
                                (datetime.datetime, np.datetime64),
                            )
                            or df[c].dtype.name.find("datetime") > -1
                        ):
                            shpfile.field(name=c, fieldType="D", size=8)
                            dfields.append(c)
                        elif isinstance(df[c].loc[idx], (bool)):
                            shpfile.field(name=c, fieldType="L", size=1)
                del c
                del idx
            for idx, row in df.iterrows():
                geom = row[df.spatial._name]
                if geom.type == "Polygon":
                    shpfile.poly(geom["rings"])
                elif geom.type == "Polyline":
                    shpfile.line(geom["paths"])
                elif geom.type == "Point":
                    shpfile.point(x=geom.x, y=geom.y)
                else:
                    shpfile.null()
                row = row[cfields].tolist()
                for fld in dfields:
                    idx = df[cfields].columns.tolist().index(fld)
                    if row[idx]:
                        if isinstance(row[idx].to_pydatetime(), (type(pd.NaT))):
                            row[idx] = None
                        else:
                            row[idx] = row[idx].to_pydatetime()
                shpfile.record(*row)
                del idx
                del row
                del geom

            # create the PRJ file
            try:
                from urllib import request

                wkid = df.spatial.sr["wkid"]
                if wkid == 102100:
                    wkid = 3857
                prj_filename = out_fc.replace(".shp", ".prj")

                url = "http://epsg.io/{}.esriwkt".format(wkid)

                opener = request.build_opener()
                opener.addheaders = [("User-Agent", "geosaurus")]
                resp = opener.open(url)

                wkt = resp.read().decode("utf-8")
                if len(wkt) > 0:
                    prj = open(prj_filename, "w")
                    prj.write(wkt)
                    prj.close()
            except:
                # Unable to write PRJ file.
                pass
            shpfile.close()
            del shpfile
            return out_fc
    return None


# --------------------------------------------------------------------------
def _pyshp2(df, out_path, out_name):
    """
    Saves a Spatially Enabled DataFrame to a Shapefile using pyshp v2.0

    :Parameters:
     :df: spatial dataframe
     :out_path: folder location to save the data
     :out_name: name of the shapefile
    :Output:
     path to the shapefile or None if pyshp isn't installed or
     spatial dataframe does not have a geometry column.
    """
    from arcgis.geometry._types import Geometry

    if HASPYSHP:
        GEOMTYPELOOKUP = {
            "Polygon": shapefile.POLYGON,
            "Point": shapefile.POINT,
            "Polyline": shapefile.POLYLINE,
            "null": shapefile.NULL,
        }

        if os.path.isdir(out_path) == False:
            os.makedirs(out_path)
        out_fc = os.path.join(out_path, out_name)
        if out_fc.lower().endswith(".shp") == False:
            out_fc += ".shp"
        geom_field = df.spatial.name
        if geom_field is None:
            return
        geom_type = "null"
        idx = df[geom_field].first_valid_index()
        if idx > -1:
            geom_type = df.loc[idx][geom_field].type
        shpfile = shapefile.Writer(
            target=out_fc,
            shapeType=GEOMTYPELOOKUP[geom_type],
            autoBalance=True,
        )

        # Start writing to shapefile
        dfields = []
        cfields = []
        for c in df.columns:
            idx = df[c].first_valid_index() or df.index.tolist()[0]
            if idx > -1:
                if isinstance(df[c].loc[idx], Geometry):
                    geom_field = (c, "GEOMETRY")
                    geom_column = c
                    # Since geometry is present, handle None type geometry occurrence
                    query_index = _handle_none_type_geometry(df, geom_type, geom_column)
                else:
                    cfields.append(c)
                    if isinstance(df[c].loc[idx], (str)):
                        shpfile.field(name=c, size=255)
                    elif isinstance(df[c].loc[idx], (int)):
                        shpfile.field(name=c, fieldType="N", size=5)
                    elif isinstance(df[c].loc[idx], np.int32):
                        shpfile.field(name=c, fieldType="N", size=10)
                    elif isinstance(df[c].loc[idx], (float, np.float64, np.int64)):
                        shpfile.field(name=c, fieldType="F", size=19, decimal=11)
                    elif (
                        isinstance(
                            df[c].loc[idx],
                            (datetime.datetime, np.datetime64),
                        )
                        or df[c].dtype.name.find("datetime") > -1
                    ):
                        shpfile.field(name=c, fieldType="D", size=8)
                        dfields.append(c)
                    elif isinstance(df[c].loc[idx], (bool)):
                        shpfile.field(name=c, fieldType="L", size=1)
            del c
            del idx

        for idx, row in df.iterrows():
            geom = row[df.spatial.name]
            if geom.type == "Polygon":
                shpfile.poly(geom["rings"])
            elif geom.type == "Polyline":
                shpfile.line(geom["paths"])
            elif geom.type == "Point":
                shpfile.point(x=geom.x, y=geom.y)
            else:
                shpfile.null()
            row = row[cfields].tolist()
            for fld in dfields:
                idx = df[cfields].columns.tolist().index(fld)
                if row[idx]:
                    if isinstance(row[idx].to_pydatetime(), (type(pd.NaT))):
                        row[idx] = None
                    else:
                        row[idx] = row[idx].to_pydatetime()
            for idx, value in enumerate(row):
                if value is np.nan:
                    row[idx] = None
            shpfile.record(*row)
            del idx
            del row
            del geom
        shpfile.close()

        # create the PRJ file
        try:
            from urllib import request

            wkid = df.spatial.sr["wkid"]
            if wkid == 102100:
                wkid = 3857
            prj_filename = out_fc.replace(".shp", ".prj")

            url = "http://epsg.io/{}.esriwkt".format(wkid)

            opener = request.build_opener()
            opener.addheaders = [("User-Agent", "geosaurus")]
            resp = opener.open(url)

            wkt = resp.read().decode("utf-8")
            if len(wkt) > 0:
                prj = open(prj_filename, "w")
                prj.write(wkt)
                prj.close()
        except:
            # Unable to write PRJ file.
            pass

        # Change back null columns to None
        for index, row in query_index.items():
            if row is True:
                df.iat[index, df.columns.get_loc(geom_column)] = None

        del shpfile
        return out_fc
    return None


def _handle_none_type_geometry(df, geom_type, geom_column):
    # Handle none type geometry occurrence

    # bool to see if empty
    query = df[geom_column].isnull()
    df_view = df[query]
    empty = df_view.empty

    if empty is False:
        for idx, row in df_view.iterrows():
            if df.loc[idx][geom_column] is None:
                if geom_type == "Point":
                    df.iat[idx, df.columns.get_loc(geom_column)] = Geometry(
                        {
                            "x": np.NAN,
                            "y": np.NAN,
                            "spatialReference": df.spatial.sr,
                        }
                    )
                elif geom_type == "Poyline":
                    df.iat[idx, df.columns.get_loc(geom_column)] = Geometry(
                        {"paths": []}
                    ).WKT
                elif geom_type == "Polygon":
                    df.iat[idx, df.columns.get_loc(geom_column)] = Geometry(
                        {"rings": []}
                    ).WKT
    return query


def _sanitize_column_names(
    geo,
    remove_special_char=True,
    rename_duplicates=True,
    inplace=False,
    use_snake_case=True,
):
    """
    Implementation for pd.DataFrame.spatial.sanitize_column_names()
    """
    original_col_names = list(geo._data.columns)

    # convert to string
    new_col_names = [str(x) for x in original_col_names]

    # use snake case
    if use_snake_case:
        import re

        for ind, val in enumerate(new_col_names):
            # skip reserved cols
            if val == geo.name:
                continue
            # replace Pascal and camel case using RE
            s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", val)
            name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()
            # remove leading spaces
            name = name.lstrip(" ")
            # replace spaces with _
            name = name.replace(" ", "_")
            # clean up too many _
            name = re.sub("_+", "_", name)
            new_col_names[ind] = name

    # remove special characters
    if remove_special_char:
        for ind, val in enumerate(new_col_names):
            name = "".join(i for i in val if i.isalnum() or "_" in i)

            # remove numeral prefixes
            for ind2, element in enumerate(name):
                if element.isdigit():
                    continue
                else:
                    name = name[ind2:]
                    break
            new_col_names[ind] = name

    # fill empty column names
    for ind, val in enumerate(new_col_names):
        if val == "":
            new_col_names[ind] = "column"

    # rename duplicates
    if rename_duplicates:
        for ind, val in enumerate(new_col_names):
            if val == geo.name:
                pass
            if new_col_names.count(val) > 1:
                counter = 1
                new_name = val + str(counter)  # adds a integer suffix to column name
                while new_col_names.count(new_name) > 0:
                    counter += 1
                    new_name = val + str(
                        counter
                    )  # if a column with the suffix exists, increment suffix
                new_col_names[ind] = new_name

    for idx, name in enumerate(new_col_names):
        if name.startswith("_"):
            new_col_names[idx] = name[1:]

    # if inplace
    if inplace:
        geo._data.columns = new_col_names
    else:
        # return a new dataframe
        df = geo._data.copy()
        df.spatial.name
        df.columns = new_col_names
        return df
    return True



import os
import sys
import traceback


HAS_BACKEND_SET = False
ARCGIS_ENABLE_TF_BACKEND = os.environ.get("ARCGIS_ENABLE_TF_BACKEND") == "1"
_LAMBDA_TEXT_CLASSIFICATION = os.environ.get("_LAMBDA_TEXT_CLASSIFICATION") == "1"

if os.environ.get("TF_CPP_MIN_LOG_LEVEL", None) is None:
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"

HAS_TENSORFLOW = False
tf_import_exception = None


class FakeImport:
    def __getattr__(self, attr):
        return self

    def __call__(self, *args, **kwargs):
        return self


if _LAMBDA_TEXT_CLASSIFICATION:
    default_module = FakeImport()
    missing_modules = [
        "scipy",
        "scipy.stats",
        "spacy",
        "spacy.symbols",
        "spacy.blank",
        "matplotlib",
        "matplotlib.pyplot",
        "matplotlib.patches",
        "matplotlib.cm",
        "scipy.special",
        "PIL",
    ]
    for module_name in missing_modules:
        sys.modules[module_name] = default_module

try:
    import platform

    if ARCGIS_ENABLE_TF_BACKEND:
        import tensorflow as tf

        HAS_TENSORFLOW = True
    elif platform.system() == "Linux":
        import tensorflow as tf
except Exception as e:
    tf_import_exception = traceback.format_exc()
    pass


def enable_backend():
    global HAS_BACKEND_SET
    global HAS_TENSORFLOW
    global ARCGIS_ENABLE_TF_BACKEND

    if ARCGIS_ENABLE_TF_BACKEND:
        # if tf.__version__ == '2.0.0':
        if HAS_TENSORFLOW and not HAS_BACKEND_SET:
            # tf.keras.backend.set_image_data_format('channels_first')
            try:
                tf.compat.v1.enable_eager_execution()
            except:
                pass
            tf_set_gpu_memory_growth()
            tf_sample_op()
            tf.keras.backend.clear_session()
            HAS_BACKEND_SET = True


def tf_set_gpu_memory_growth():
    gpus = tf.config.experimental.list_physical_devices("GPU")
    if gpus:
        try:
            # Currently, memory growth needs to be the same across GPUs
            for gpu in gpus:
                tf.config.experimental.set_memory_growth(gpu, True)
            logical_gpus = tf.config.experimental.list_logical_devices("GPU")
            # print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
        except RuntimeError as e:
            # Memory growth must be set before GPUs have been initialized
            print(e)


def raise_tensorflow_import_error():
    if ARCGIS_ENABLE_TF_BACKEND:
        message = """
        Could not find tensorflow, Please install tensorflow using the following command
        \nconda install -c esri tensorflow-gpu=2.1.0
        """
        ex = Exception(message)
        raise (ex)
    else:
        message = """
        Please enable tensorflow by setting the required environment variable 'ARCGIS_ENABLE_TF_BACKEND' to '1' before importing arcgis
        \n for example the following code block needs to be executed before importing arcgis
        \n\n`import os; os.environ['ARCGIS_ENABLE_TF_BACKEND'] = '1'`
        """
        ex = Exception(message)
        raise (ex)


def tf_sample_op():
    a = tf.keras.layers.Conv2D(1, (3, 3))
    a = a(tf.zeros((1, 3, 20, 20))).numpy()


## Fastai Imports #######

HAS_FASTAI = False
fastai_import_exception = None


def do_fastai_imports():
    global HAS_FASTAI
    global fastai_import_exception

    try:
        import fastai
        import torch
        import torchvision
        import skimage

        HAS_FASTAI = True
    except Exception as e:
        fastai_import_exception = traceback.format_exc()
        pass

    try:
        from .patches import precondition
    except:
        pass


def fastai_installation_command():
    installation_steps = "Install them using 'conda install -c esri arcgis=1.8.1 pillow scikit-image'\n'conda install -c fastai -c pytorch fastai pytorch=1.4.0 torchvision=0.5.0 tensorflow-gpu=2.1.0'\n'conda install gdal=2.3.3'"

    return installation_steps


def raise_fastai_import_error(
    import_exception=fastai_import_exception, installation_steps=None, message=None
):
    if installation_steps is None:
        installation_steps = fastai_installation_command()
    if message is None:
        message = "This module requires fastai, PyTorch, torchvision and scikit-image as its dependencies."
    raise Exception(f"""{import_exception} \n\n{message}\n{installation_steps}""")


HAS_GDAL = False
gdal_import_exception = None
GDAL_INSTALL_MESSAGE = f"""
\nPlease install gdal using the following command
\nconda install gdal=2.3.3
""".strip()

try:
    from osgeo import gdal

    HAS_GDAL = True
except Exception as e:
    gdal_import_exception = traceback.format_exc()
    pass


def raise_gdal_import_error(import_exception=gdal_import_exception):
    message = "gdal is required to work with multispectral datasets."
    raise Exception(f"""{import_exception} \n\n{message}\n{GDAL_INSTALL_MESSAGE}""")


## Ipython inside ArcGIS Pro
import sys

_IS_ARCGISPRONOTEBOOK = None
using_mpl_inline = False


def is_arcgispronotebook():
    global using_mpl_inline
    if _IS_ARCGISPRONOTEBOOK is not None:
        if _IS_ARCGISPRONOTEBOOK:
            patch_arcgis_notebook()
        else:
            using_mpl_inline = False
        return _IS_ARCGISPRONOTEBOOK
    if os.path.basename(sys.executable) == "ArcGISPro.exe":
        ### This code will be execute only once in ArcGIS Pro notebooks
        if not using_mpl_inline:
            patch_arcgis_notebook()
            using_mpl_inline = True
        ###
        ### This following code block will be replaced by a flag exposed by Pro's Team (Vinay Vijayan)
        return True
        ###


def reload_IPython():
    if "IPython" in sys.modules:
        del sys.modules["IPython"]
    import IPython

    return IPython


def patch_arcgis_notebook():
    if reload_IPython().get_ipython() is not None:
        reload_IPython().get_ipython().run_line_magic("matplotlib", "inline")


is_arcgispronotebook()



import numpy as np

from arcgis.auth.tools import LazyLoader

cv2 = LazyLoader("cv2")


def _pad_image(image, stride):
    img_h = image.shape[0] + stride * 2
    img_w = image.shape[1] + stride * 2

    tmp = np.zeros((img_h, img_w, image.shape[2]))
    tmp[stride : stride + image.shape[0], stride : stride + image.shape[1], :] = image[
        :, :, :
    ]
    img = tmp.astype(np.uint8)

    return img


def _get_image_chips(image, chip_dim):
    """Function to take an image and
    return sequentially cropped and padded chips of size chip_dim."""

    img_h, img_w, _ = image.shape
    chips_data = []
    stride = chip_dim // 2

    start_x = 0
    start_y = 0

    if chip_dim >= img_w and chip_dim >= img_h:
        return [
            {
                "height": img_h,
                "width": img_w,
                "chip": image,
                "xmin": start_x,
                "ymin": start_y,
                "predictions": [],
            }
        ]

    # Add zero-padding to enable prediction on all parts of the image
    padded_image = _pad_image(image, stride)
    img_h, img_w, _ = padded_image.shape

    while start_x + chip_dim <= img_w:
        start_y = 0
        while start_y + chip_dim <= img_h:
            chip = padded_image[
                start_y : start_y + chip_dim, start_x : start_x + chip_dim, :
            ]

            if (chip.shape[0] != chip_dim) or (chip.shape[1] != chip_dim):
                tmp = np.zeros((chip_dim, chip_dim, 3))
                tmp[0 : chip.shape[0], 0 : chip.shape[1], :] = chip[:, :, :]
                chip = tmp.astype(np.uint8)

            # To translate the bbox coordinates according to unpadded image
            xmin = start_x - stride
            ymin = start_y - stride

            chips_data.append(
                {
                    "height": chip_dim,
                    "width": chip_dim,
                    "chip": chip,
                    "xmin": xmin,
                    "ymin": ymin,
                    "predictions": [],
                }
            )
            start_y = start_y + stride

        start_x = start_x + stride

    return chips_data


def _get_transformed_predictions(chips_data):
    predictions = []
    labels = []
    scores = []
    for chip_data in chips_data:
        for prediction in chip_data["predictions"]:
            prediction["xmin"] = prediction["xmin"] + chip_data["xmin"]
            prediction["ymin"] = prediction["ymin"] + chip_data["ymin"]

            predictions.append(
                [
                    prediction["xmin"],
                    prediction["ymin"],
                    prediction["width"],
                    prediction["height"],
                ]
            )
            labels.append(prediction["label"].obj)
            scores.append(prediction["score"])

    return predictions, labels, scores


def _draw_predictions(
    frame,
    predictions,
    labels,
    scores=None,
    show_scores=True,
    show_labels=True,
    color=(255, 255, 255),
    fontface=0,
    thickness=2,
):
    for index, data in enumerate(predictions):
        frame = cv2.rectangle(
            frame,
            (int(data[0]), int(data[1])),
            (int(data[0] + data[2]), int(data[1] + data[3])),
            color,
            thickness,
        )

        text_to_display = None
        if show_labels:
            text_to_display = str(labels[index])

        if show_scores and scores is not None:
            if text_to_display:
                scores[index] = float("{0:.2f}".format(scores[index]))
                text_to_display = text_to_display + ": " + str(scores[index])
            else:
                text_to_display = str(scores[index])

        if text_to_display:
            frame = cv2.putText(
                frame,
                text_to_display,
                (int(data[0]), int(data[1]) - 10),
                fontface,
                0.7,
                color,
                thickness,
            )

    return frame


def _exclude_detection(data, chip_width, chip_height):
    if chip_height < chip_width:
        padding = chip_height // 4
    else:
        padding = chip_width // 4

    center_coord_x = data[0] + data[2] / 2
    center_coord_y = data[1] + data[3] / 2

    if (
        center_coord_x < padding
        or center_coord_y < padding
        or center_coord_x > (chip_width - padding)
        or center_coord_y > (chip_height - padding)
    ):
        return True

    return False
